import shutil
import tempfile
from sys import stderr
from typing import (
    Any,
    Optional,
)

from exasol_integration_test_docker_environment.lib.api import (
    spawn_test_environment,
    spawn_test_environment_with_test_container,
)
from exasol_integration_test_docker_environment.lib.models.data.test_container_content_description import (
    TestContainerContentDescription,
)
from exasol_integration_test_docker_environment.lib.test_environment.ports import Ports
from exasol_integration_test_docker_environment.testing.docker_registry import (
    default_docker_repository_name,
)
from exasol_integration_test_docker_environment.testing.exaslct_docker_test_environment import (
    ExaslctDockerTestEnvironment,
)
from exasol_integration_test_docker_environment.testing.exaslct_test_environment import (
    get_class,
    get_test_flavor,
)


class ApiTestEnvironment:

    def __init__(self, test_object, name=None) -> None:
        self.test_object = test_object
        self.test_class = get_class(test_object)
        self.flavor_path = get_test_flavor(self.test_class)
        self.name = name if name else self.test_class.__name__
        self.docker_repository_name = default_docker_repository_name(self.name)
        self.temp_dir = tempfile.mkdtemp()

    @property
    def output_dir(self):
        return self.temp_dir

    @property
    def task_dependency_dot_file(self):
        return f"{self.name}.dot"

    def close(self):
        try:
            shutil.rmtree(self.temp_dir)
        except Exception as e:
            print(e, file=stderr)

    def _get_default_test_environment(self, name: str, ports: Ports):
        return ExaslctDockerTestEnvironment(
            name=self.name + "_" + name,
            database_host="localhost",
            db_username="sys",
            db_password="exasol",
            bucketfs_username="w",
            bucketfs_password="write",
            ports=ports,
        )

    def spawn_docker_test_environment_with_test_container(
        self,
        name: str,
        test_container_content: TestContainerContentDescription,
        additional_parameter: Optional[dict[str, Any]] = None,
    ) -> ExaslctDockerTestEnvironment:
        if additional_parameter is None:
            additional_parameter = {}
        ports = Ports.random_free()
        on_host_parameter = self._get_default_test_environment(name, ports)
        docker_db_image_version = on_host_parameter.docker_db_image_version
        on_host_parameter.environment_info, on_host_parameter.clean_up = (
            spawn_test_environment_with_test_container(
                environment_name=on_host_parameter.name,
                database_port_forward=ports.database,
                bucketfs_http_port_forward=ports.bucketfs_http,
                bucketfs_https_port_forward=ports.bucketfs_https,
                ssh_port_forward=ports.ssh,
                docker_db_image_version=docker_db_image_version,
                test_container_content=test_container_content,
                output_directory=self.output_dir,
                **additional_parameter,
            )
        )
        return on_host_parameter

    def spawn_docker_test_environment(
        self,
        name: str,
        additional_parameter: Optional[dict[str, Any]] = None,
    ) -> ExaslctDockerTestEnvironment:
        if additional_parameter is None:
            additional_parameter = {}

        ports = Ports.random_free()
        if "bucketfs_port_forward" in additional_parameter:
            ports.bucketfs_http = additional_parameter["bucketfs_port_forward"]
            del additional_parameter["bucketfs_port_forward"]
        if "bucketfs_https_port_forward" in additional_parameter:
            ports.bucketfs_https = additional_parameter["bucketfs_https_port_forward"]
            del additional_parameter["bucketfs_https_port_forward"]

        on_host = self._get_default_test_environment(name, ports)
        on_host.environment_info, on_host.clean_up = spawn_test_environment(
            environment_name=on_host.name,
            database_port_forward=on_host.ports.database,
            bucketfs_port_forward=on_host.ports.bucketfs_http,
            bucketfs_https_port_forward=on_host.ports.bucketfs_https,
            ssh_port_forward=on_host.ports.ssh,
            docker_db_image_version=on_host.docker_db_image_version,
            output_directory=self.output_dir,
            **additional_parameter,
        )
        return on_host



import click

from exasol_integration_test_docker_environment.lib.test_environment.parameter.docker_db_test_environment_parameter import (
    DbOsAccess,
)
from exasol_integration_test_docker_environment.lib.test_environment.ports import Ports

test_environment_options = [
    click.option(
        "--environment-type",
        type=click.Choice(["docker_db", "external_db"]),
        default="""docker_db""",
        show_default=True,
        help="""Environment type for tests.""",
    ),
    click.option(
        "--max_start_attempts",
        type=int,
        default=2,
        show_default=True,
        help="""Maximum start attempts for environment""",
    ),
]

LATEST_DB_VERSION = """2025.1.8"""
DEFAULT_MEM_SIZE = """2 GiB"""
DEFAULT_DISK_SIZE = """2 GiB"""

docker_db_options = [
    click.option(
        "--docker-db-image-version",
        type=str,
        default=LATEST_DB_VERSION,
        show_default=True,
        help="""Docker DB Image Version against which the tests should run.""",
    ),
    click.option(
        "--docker-db-image-name",
        type=str,
        default="""exasol/docker-db""",
        show_default=True,
        help="""Docker DB Image Name against which the tests should run.""",
    ),
    click.option(
        "--db-os-access",
        type=click.Choice([e.name for e in DbOsAccess]),
        metavar="METHOD",
        default="""DOCKER_EXEC""",
        show_default=True,
        help="""How to access file system and command line of the
     		 database operating system. Experimental option, will show no
     		 effect until implementation of feature SSH access is
     		 completed.""",
    ),
    click.option(
        "--create-certificates/--no-create-certificates",
        default=False,
        help="""Creates and injects SSL certificates to the Docker DB container.""",
    ),
    click.option(
        "--additional-db-parameter",
        "-p",
        type=str,
        multiple=True,
        default=(),
        help="""Additional database parameter which will be injected to EXAConf. Value should have format '-param=value'.""",
    ),
    click.option(
        "--docker-environment-variable",
        type=str,
        multiple=True,
        default=[],
        help="""An environment variable which will be added to the docker-db.
                The variable needs to have format "key=value".
                For example "HTTPS_PROXY=192.168.1.5".
                You can repeat this option to add further environment variables.""",
    ),
    click.option(
        "--accelerator",
        multiple=True,
        default=[],
        help="""Configures the nvidia container toolkit to use the given GPU. 
                Currently only value 'nvidia=all' is supported. 
                For 2025.1.x and later only""",
    ),
]

external_db_options = [
    click.option(
        "--external-exasol-db-host",
        type=str,
        help="""Host name or IP of external Exasol DB, needs to be set if --environment-type=external_db""",
    ),
    click.option(
        "--external-exasol-db-port",
        type=int,
        default=Ports.external.database,
        help="""Database port of external Exasol DB, needs to be set if --environment-type=external_db""",
    ),
    click.option(
        "--external-exasol-bucketfs-port",
        type=int,
        default=Ports.external.bucketfs_http,
        help="""Bucketfs port of external Exasol DB, needs to be set if --environment-type=external_db""",
    ),
    click.option(
        "--external-exasol-ssh-port",
        type=int,
        help="""SSH port of external Exasol DB, needs to be set if --environment-type=external_db""",
    ),
    click.option(
        "--external-exasol-db-user",
        type=str,
        help="""User for external Exasol DB, needs to be set if --environment-type=external_db""",
    ),
    click.option(
        "--external-exasol-db-password",
        type=str,
        help="""Database Password for external Exasol DB""",
    ),
    click.option(
        "--external-exasol-bucketfs-write-password",
        type=str,
        help="""BucketFS write Password for external Exasol DB""",
    ),
    click.option(
        "--external-exasol-xmlrpc-host",
        type=str,
        help="""Hostname for the xmlrpc server""",
    ),
    click.option(
        "--external-exasol-xmlrpc-port",
        type=int,
        default=443,
        show_default=True,
        help="""Port for the xmlrpc server""",
    ),
    click.option(
        "--external-exasol-xmlrpc-user",
        type=str,
        default="""admin""",
        show_default=True,
        help="""User for the xmlrpc server""",
    ),
    click.option(
        "--external-exasol-xmlrpc-password",
        type=str,
        help="""Password for the xmlrpc server""",
    ),
    click.option(
        "--external-exasol-xmlrpc-cluster-name",
        type=str,
        default="""cluster1""",
        show_default=True,
        help="""Password for the xmlrpc server""",
    ),
]



import click

DEFAULT_OUTPUT_DIRECTORY = ".build_output"

output_directory_option = click.option(
    "--output-directory",
    type=click.Path(file_okay=False, dir_okay=True),
    default=DEFAULT_OUTPUT_DIRECTORY,
    show_default=True,
    help="Output directory where the system stores all output and log files.",
)

tempory_base_directory_option = click.option(
    "--temporary-base-directory",
    type=click.Path(file_okay=False, dir_okay=True),
    default="/tmp",
    show_default=True,
    help="Directory where the system creates temporary directories.",
)

system_options = [
    click.option(
        "--workers",
        type=int,
        default=5,
        show_default=True,
        help="Number of parallel workers",
    ),
    click.option(
        "--task-dependencies-dot-file",
        type=click.Path(file_okay=True),
        default=None,
        help="Path where to store the Task Dependency Graph as dot file",
    ),
]

luigi_logging_options = [
    click.option(
        "--log-level",
        type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "FATAL"]),
        default=None,
        help="Log level used for console logging",
    ),
    click.option(
        "--use-job-specific-log-file",
        type=bool,
        default=True,
        help="Use a job specific log file which write the debug log "
        "to the job directory in the build directory",
    ),
]



import contextlib
from collections.abc import (
    Generator,
    Iterator,
)
from typing import (
    Any,
    ContextManager,
    Optional,
    Protocol,
)

from exasol_integration_test_docker_environment.lib.models.data.test_container_content_description import (
    TestContainerContentDescription,
)
from exasol_integration_test_docker_environment.testing import (
    luigi_utils,
    utils,
)
from exasol_integration_test_docker_environment.testing.api_test_environment import (
    ApiTestEnvironment,
)
from exasol_integration_test_docker_environment.testing.exaslct_docker_test_environment import (
    ExaslctDockerTestEnvironment,
)


@contextlib.contextmanager
def build_api_isolation(request) -> Iterator[ApiTestEnvironment]:
    """
    Builds an ApiTestEnvironment instance with a proper name based on the pytest request fixture.
    Cleans up the environment automatically on shutdown.
    """
    testname = utils.normalize_request_name(request.node.name)
    environment = ApiTestEnvironment(test_object=None, name=testname)
    yield environment
    utils.close_environments(environment)
    luigi_utils.clean(environment.docker_repository_name)


class ApiContextProvider(Protocol):
    def __call__(
        self,
        name: Optional[str],
        additional_parameters: Optional[dict[str, Any]] = None,
    ) -> ContextManager[ExaslctDockerTestEnvironment]: ...


def build_api_context_provider(
    test_environment: ApiTestEnvironment,
) -> ApiContextProvider:
    """
    Returns a context provider function which can be used to spawn a Docker DB with custom name and custom additional db parameters.
    Cleans up the docker db automatically on shutdown when context provider closes.
    """

    @contextlib.contextmanager
    def create_context(
        name: Optional[str] = None,
        additional_parameters: Optional[dict[str, Any]] = None,
    ) -> Generator[ExaslctDockerTestEnvironment, None, None]:
        name = name if name else test_environment.name
        spawned = test_environment.spawn_docker_test_environment(
            name=name,
            additional_parameter=additional_parameters,
        )
        yield spawned
        utils.close_environments(spawned)

    return create_context


class ApiContextProviderWithTestContainer(Protocol):
    def __call__(
        self,
        name: Optional[str],
        additional_parameters: Optional[dict[str, Any]] = None,
        test_container_content: Optional[TestContainerContentDescription] = None,
    ) -> ContextManager[ExaslctDockerTestEnvironment]: ...


def build_api_context_provider_with_test_container(
    test_environment: ApiTestEnvironment,
    default_test_container_content: TestContainerContentDescription,
) -> ApiContextProviderWithTestContainer:
    """
    Returns a context provider function which can be used to spawn a Docker DB + Test container
    with custom name, custom additional db parameters and custom test container content.
    Cleans up the docker db + test container automatically on shutdown when context provider closes.
    """

    @contextlib.contextmanager
    def create_context(
        name: Optional[str] = None,
        additional_parameters: Optional[dict[str, Any]] = None,
        test_container_content=default_test_container_content,
    ) -> Generator[ExaslctDockerTestEnvironment, None, None]:
        name = name if name else test_environment.name
        spawned = test_environment.spawn_docker_test_environment_with_test_container(
            name=name,
            additional_parameter=additional_parameters,
            test_container_content=test_container_content,
        )
        yield spawned
        utils.close_environments(spawned)

    return create_context



import json

from exasol_integration_test_docker_environment.lib.docker.images.image_info import (
    ImageInfo,
)
from exasol_integration_test_docker_environment.lib.logging.abstract_log_handler import (
    AbstractLogHandler,
)
from exasol_integration_test_docker_environment.lib.models.config.log_config import (
    WriteLogFilesToConsole,
)


class PullLogHandler(AbstractLogHandler):

    def __init__(self, log_file_path, logger, image_info: ImageInfo) -> None:
        super().__init__(log_file_path, logger)
        self._image_info = image_info

    def handle_log_line(self, log_line, error: bool = False) -> None:
        json_output = json.loads(log_line)
        if (
            "status" in json_output
            and json_output["status"] != "Downloading"
            and json_output["status"] != "Extracting"
        ):
            self._complete_log.append(json_output["status"])
            self._log_file.write(json_output["status"])
            self._log_file.write("\n")
            self._log_file.flush()
        if "errorDetail" in json_output:
            self._error_message = json_output["errorDetail"]["message"]

    def finish(self) -> None:
        self.write_log_to_console_if_requested()
        if self._error_message is not None:
            self.write_error_log_to_console_if_requested()
            raise Exception(
                'Error occurred during the pull of the image %s. Received error "%s" .'
                "The whole log can be found in %s"
                % (
                    self._image_info.get_source_complete_name(),
                    self._error_message,
                    self._log_file_path,
                )
            )

    def write_error_log_to_console_if_requested(self) -> None:
        if (
            self._log_config.write_log_files_to_console
            == WriteLogFilesToConsole.only_error
        ):
            self._logger.error(
                "Pull of image %s failed\nPush Log:\n%s",
                self._image_info.get_source_complete_name(),
                "\n".join(self._complete_log),
            )

    def write_log_to_console_if_requested(self) -> None:
        if self._log_config.write_log_files_to_console == WriteLogFilesToConsole.all:
            self._logger.info(
                "Pull Log of image %s\n%s",
                self._image_info.get_source_complete_name(),
                "\n".join(self._complete_log),
            )



import shutil
import tempfile
from pathlib import Path

from exasol_integration_test_docker_environment.lib.base.still_running_logger import (
    StillRunningLogger,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_creator_base_task import (
    DockerImageCreatorBaseTask,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.build_context_creator import (
    BuildContextCreator,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.build_log_handler import (
    BuildLogHandler,
)
from exasol_integration_test_docker_environment.lib.docker.images.image_info import (
    ImageInfo,
)
from exasol_integration_test_docker_environment.lib.models.config.build_config import (
    build_config,
)
from exasol_integration_test_docker_environment.lib.models.config.docker_config import (
    docker_build_arguments,
)


class DockerBuildImageTask(DockerImageCreatorBaseTask):

    def run_task(self) -> None:
        self.logger.info(
            "Build docker image %s, log file can be found here %s",
            self.image_info.get_target_complete_name(),
            self.get_log_path(),
        )
        temp_directory = tempfile.mkdtemp(
            prefix="script_language_container_tmp_dir",
            dir=build_config().temporary_base_directory,
        )
        try:
            image_description = self.image_info.image_description
            build_context_creator = BuildContextCreator(
                temp_directory, self.image_info, self.get_log_path()
            )
            build_context_creator.prepare_build_context_to_temp_dir()
            assert image_description is not None
            with self._get_docker_client() as docker_client:
                output_generator = docker_client.api.build(
                    path=temp_directory,
                    nocache=self.no_cache,
                    tag=self.image_info.get_target_complete_name(),
                    rm=True,
                    buildargs=dict(
                        **image_description.transparent_build_arguments,
                        **image_description.image_changing_build_arguments,
                        **docker_build_arguments().secret
                    ),
                )
                self._handle_output(output_generator, self.image_info)
        finally:
            shutil.rmtree(temp_directory)

    def _handle_output(self, output_generator, image_info: ImageInfo):
        log_file_path = Path(self.get_log_path(), "docker-build.log")
        with BuildLogHandler(log_file_path, self.logger, image_info) as log_handler:
            still_running_logger = StillRunningLogger(
                self.logger, "build image %s" % image_info.get_target_complete_name()
            )
            for log_line in output_generator:
                still_running_logger.log()
                log_handler.handle_log_lines(log_line)



from typing import Optional

from exasol_integration_test_docker_environment.lib.base.db_os_executor import (
    DbOsExecFactory,
    DockerClientFactory,
    DockerExecFactory,
    SshExecFactory,
)
from exasol_integration_test_docker_environment.lib.models.data.database_info import (
    DatabaseInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.docker_network_info import (
    DockerNetworkInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.docker_volume_info import (
    DockerVolumeInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.environment_type import (
    EnvironmentType,
)
from exasol_integration_test_docker_environment.lib.test_environment.abstract_spawn_test_environment import (
    AbstractSpawnTestEnvironment,
)
from exasol_integration_test_docker_environment.lib.test_environment.create_certificates.create_ssl_certificates_task import (
    CreateSSLCertificatesTask,
)
from exasol_integration_test_docker_environment.lib.test_environment.database_waiters.wait_for_test_docker_database import (
    WaitForTestDockerDatabase,
)
from exasol_integration_test_docker_environment.lib.test_environment.db_version import (
    db_version_supports_custom_certificates,
)
from exasol_integration_test_docker_environment.lib.test_environment.parameter.docker_db_test_environment_parameter import (
    DbOsAccess,
    DockerDBTestEnvironmentParameter,
)
from exasol_integration_test_docker_environment.lib.test_environment.prepare_network_for_test_environment import (
    PrepareDockerNetworkForTestEnvironment,
)
from exasol_integration_test_docker_environment.lib.test_environment.spawn_test_database import (
    SpawnTestDockerDatabase,
)


class SpawnTestEnvironmentWithDockerDB(
    AbstractSpawnTestEnvironment, DockerDBTestEnvironmentParameter
):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.db_container_name = f"""db_container_{self.environment_name}"""
        self.certificate_volume_name = f"""certificates_{self.environment_name}"""

    def get_environment_type(self):
        return EnvironmentType.docker_db

    def create_ssl_certificates(self):
        if not db_version_supports_custom_certificates(self.docker_db_image_version):
            raise ValueError(
                "Minimal supported Database with custom certificates is '7.0.6'"
            )
        return self.create_child_task_with_common_params(
            CreateSSLCertificatesTask,
            environment_name=self.environment_name,
            db_container_name=self.db_container_name,
            network_name=self.network_name,
            docker_runtime=self.docker_runtime,
            volume_name=self.certificate_volume_name,
            reuse=self.reuse_database or self.reuse_test_container,
            no_cleanup_after_success=self.no_database_cleanup_after_success
            or self.no_test_container_cleanup_after_success,
            no_cleanup_after_failure=self.no_database_cleanup_after_failure
            or self.no_test_container_cleanup_after_failure,
        )

    def create_network_task(self, attempt: int):
        return self.create_child_task_with_common_params(
            PrepareDockerNetworkForTestEnvironment,
            test_container_name=self.test_container_name,
            network_name=self.network_name,
            db_container_name=self.db_container_name,
            reuse=self.reuse_database or self.reuse_test_container,
            no_cleanup_after_success=self.no_database_cleanup_after_success
            or self.no_test_container_cleanup_after_success,
            no_cleanup_after_failure=self.no_database_cleanup_after_failure
            or self.no_test_container_cleanup_after_failure,
            attempt=attempt,
        )

    def _executor_factory(self, database_info: DatabaseInfo) -> DbOsExecFactory:
        if self.db_os_access == DbOsAccess.SSH:
            return SshExecFactory.from_database_info(database_info)
        client_factory = DockerClientFactory(timeout=100000)
        return DockerExecFactory(self.db_container_name, client_factory)

    def create_spawn_database_task(
        self,
        network_info: DockerNetworkInfo,
        certificate_volume_info: Optional[DockerVolumeInfo],
        attempt: int,
    ):
        def volume_name(info):
            return None if info is None else info.volume_name

        return self.create_child_task_with_common_params(
            SpawnTestDockerDatabase,
            db_container_name=self.db_container_name,
            network_info=network_info,
            certificate_volume_name=volume_name(certificate_volume_info),
            ip_address_index_in_subnet=0,
            attempt=attempt,
            additional_db_parameter=self.additional_db_parameter,
            docker_environment_variables=self.docker_environment_variables,
        )

    def create_wait_for_database_task(
        self, attempt: int, database_info: DatabaseInfo
    ) -> WaitForTestDockerDatabase:
        return self.create_child_task_with_common_params(
            WaitForTestDockerDatabase,
            database_info=database_info,
            attempt=attempt,
            docker_db_image_version=self.docker_db_image_version,
            executor_factory=self._executor_factory(database_info),
        )



from pathlib import Path

from exasol_integration_test_docker_environment.lib import (
    extract_modulename_for_build_steps,
)


def import_build_steps(flavor_path: tuple[str, ...]) -> None:
    # TODO Move to script-languages-container-tools: https://github.com/exasol/script-languages-container-tool/issues/268
    # We need to load the build steps of a flavor in the commandline processor,
    # because the imported classes need to be available in all processes spawned by luigi.
    # If we  import the build steps in a Luigi Task they are only available in the worker
    # which executes this task. The build method with local scheduler of luigi uses fork
    # to create the scheduler and worker processes, such that the imported classes available
    # in the scheduler and worker processes
    import importlib.util

    for path in flavor_path:
        path_to_build_steps = Path(path).joinpath("flavor_base/build_steps.py")
        module_name_for_build_steps = extract_modulename_for_build_steps(path)
        spec = importlib.util.spec_from_file_location(
            module_name_for_build_steps, path_to_build_steps
        )
        assert spec and spec.loader
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)



import logging


class TaskLoggerWrapper:
    def __init__(self, logger: logging.Logger, task_id) -> None:
        self.task_id = task_id
        self.logger = logger

    def get_message(self, msg):
        return f"{self.task_id}: {msg}"

    def debug(self, msg, *args, **kwargs):
        self.logger.debug(self.get_message(msg), *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.logger.info(self.get_message(msg), *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.logger.warning(self.get_message(msg), *args, **kwargs)

    def warn(self, msg, *args, **kwargs):
        self.logger.warn(self.get_message(msg), *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.logger.error(self.get_message(msg), *args, **kwargs)

    def exception(self, msg, *args, exc_info=True, **kwargs):
        self.logger.exception(self.get_message(msg), *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.logger.critical(self.get_message(msg), *args, **kwargs)

    fatal = critical



import time

from exasol_integration_test_docker_environment.lib.test_environment.database_setup.bucketfs_sync_checker import (
    BucketFSSyncChecker,
)


class TimeBasedBucketFSSyncWaiter(BucketFSSyncChecker):

    def __init__(self, sync_time_estimation_in_seconds: float) -> None:
        self._sync_time_estimation_in_seconds = sync_time_estimation_in_seconds

    def prepare_upload(self):
        pass

    def wait_for_bucketfs_sync(self):
        time.sleep(self._sync_time_estimation_in_seconds)



import time
from pathlib import PurePath
from threading import Thread

from docker.models.containers import Container

from exasol_integration_test_docker_environment.lib.base.db_os_executor import (
    DbOsExecFactory,
)
from exasol_integration_test_docker_environment.lib.base.task_logger_wrapper import (
    TaskLoggerWrapper,
)
from exasol_integration_test_docker_environment.lib.models.data.database_credentials import (
    DatabaseCredentials,
)
from exasol_integration_test_docker_environment.lib.models.data.database_info import (
    DatabaseInfo,
)
from exasol_integration_test_docker_environment.lib.test_environment.database_setup.find_exaplus_in_db_container import (
    find_exaplus,
)


class IsDatabaseReadyThread(Thread):
    def __init__(
        self,
        logger: TaskLoggerWrapper,
        database_info: DatabaseInfo,
        database_container: Container,
        database_credentials: DatabaseCredentials,
        docker_db_image_version: str,
        executor_factory: DbOsExecFactory,
    ) -> None:
        super().__init__()
        self.logger = logger
        self.database_credentials = database_credentials
        self._database_info = database_info
        self._db_container = database_container
        self.finish = False
        self.is_ready = False
        self.output_db_connection = None
        self.output_bucketfs_connection = None
        self.docker_db_image_version = docker_db_image_version
        self.executor_factory = executor_factory

    def stop(self):
        self.logger.info("Stop IsDatabaseReadyThread")
        self.finish = True

    def run(self):
        try:
            with self.executor_factory.executor() as executor:
                db_connection_command = ""
                bucket_fs_connection_command = ""
                try:
                    executor.prepare()
                    exaplus_path = find_exaplus(self._db_container, executor)
                    db_connection_command = self.create_db_connection_command(
                        exaplus_path
                    )
                    bucket_fs_connection_command = (
                        self.create_bucketfs_connection_command()
                    )
                except RuntimeError as e:
                    self.logger.exception(
                        "Caught exception while searching for exaplus."
                    )
                    self.finish = True
                while not self.finish:
                    (exit_code_db_connection, self.output_db_connection) = (
                        executor.exec(db_connection_command)
                    )
                    (exit_code_bucketfs_connection, self.output_bucketfs_connection) = (
                        executor.exec(bucket_fs_connection_command)
                    )
                    if (
                        exit_code_db_connection == 0
                        and exit_code_bucketfs_connection == 0
                    ):
                        self.finish = True
                        self.is_ready = True
                    time.sleep(1)
        except Exception as e:
            self.finish = True
            self.logger.exception("Caught exception in IsDatabaseReadyThread.run.")

    def create_db_connection_command(self, exaplus_path: PurePath):
        username = self.database_credentials.db_user
        password = self.database_credentials.db_password
        connection_options = f"""-c 'localhost:{self._database_info.ports.database}' -u '{username}' -p '{password}'"""

        cmd = f"""{exaplus_path} {connection_options} -sql 'select 1;' -jdbcparam 'validateservercertificate=0'"""
        bash_cmd = f"""bash -c "{cmd}" """
        return bash_cmd

    def create_bucketfs_connection_command(self):
        username = "w"
        password = self.database_credentials.bucketfs_write_password
        cmd = f"""curl --silent --show-error --fail '{username}:{password}@localhost:{self._database_info.ports.bucketfs_http}'"""
        bash_cmd = f"""bash -c "{cmd}" """
        return bash_cmd



import json

import jsonpickle

from exasol_integration_test_docker_environment.lib.base.frozendict_to_dict import (
    FrozenDictToDict,
)


class Info:
    def to_json(self, indent=4) -> str:
        jsonpickle.set_preferred_backend("simplejson")
        jsonpickle.set_encoder_options("simplejson", sort_keys=True, indent=indent)
        return jsonpickle.encode(self)

    def to_dict(self) -> dict:
        return json.loads(self.to_json())

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__dict__})"

    @classmethod
    def from_json(cls, json_string: str):
        loaded_object = jsonpickle.decode(json_string)
        if not isinstance(loaded_object, cls):
            raise TypeError(
                "Type {} of loaded object does not match {}".format(
                    type(loaded_object), cls
                )
            )
        return loaded_object

    @classmethod
    def from_dict(cls, dictionary: dict):
        converted_dictionary = FrozenDictToDict().convert(dictionary)
        dumps = json.dumps(converted_dictionary)
        return cls.from_json(dumps)



import contextlib
import warnings
from collections.abc import Iterator
from pathlib import Path
from typing import (
    Optional,
)

import luigi
from luigi.parameter import UnconsumedParameterWarning
from luigi.setup_logging import InterfaceLogging

from exasol_integration_test_docker_environment.lib.logging.luigi_log_config import (
    get_luigi_log_config,
)


@contextlib.contextmanager
def configure_logging(
    log_file_path: Path, log_level: Optional[str], use_job_specific_log_file: bool
) -> Iterator[dict[str, str]]:
    with get_luigi_log_config(
        log_file_target=log_file_path,
        log_level=log_level,
        use_job_specific_log_file=use_job_specific_log_file,
    ) as luigi_config:
        no_configure_logging, run_kwargs = _configure_logging_parameter(
            log_level=log_level,
            luigi_config=luigi_config,
            use_job_specific_log_file=use_job_specific_log_file,
        )
        # We need to set InterfaceLogging._configured to false,
        # because otherwise luigi doesn't accept the new config.
        InterfaceLogging._configured = False
        luigi.configuration.get_config().set(
            "core", "no_configure_logging", str(no_configure_logging)
        )
        with warnings.catch_warnings():
            # This filter is necessary, because luigi uses the config no_configure_logging,
            # but doesn't define it, which seems to be a bug in luigi
            warnings.filterwarnings(
                action="ignore",
                category=UnconsumedParameterWarning,
                message=".*no_configure_logging.*",
            )
            yield run_kwargs


def _configure_logging_parameter(
    log_level: Optional[str], luigi_config: Path, use_job_specific_log_file: bool
) -> tuple[bool, dict[str, str]]:
    if use_job_specific_log_file:
        no_configure_logging = False
        run_kwargs = {"logging_conf_file": f"{luigi_config}"}
    else:
        no_configure_logging = True
        run_kwargs = {}
    return no_configure_logging, run_kwargs



from typing import (
    Optional,
)

import luigi

from exasol_integration_test_docker_environment.lib.test_environment.parameter.test_container_parameter import (
    OptionalTestContainerParameter,
)


class GeneralSpawnTestEnvironmentParameter(OptionalTestContainerParameter):
    reuse_database_setup: bool = luigi.BoolParameter(False, significant=False)
    reuse_test_container: bool = luigi.BoolParameter(False, significant=False)
    no_test_container_cleanup_after_success: bool = luigi.BoolParameter(
        False, significant=False
    )
    no_test_container_cleanup_after_failure: bool = luigi.BoolParameter(
        False, significant=False
    )
    max_start_attempts: int = luigi.IntParameter(2, significant=False)
    docker_runtime: Optional[str] = luigi.OptionalParameter(None, significant=False)
    create_certificates: bool = luigi.BoolParameter()
    additional_db_parameter: tuple[str, ...] = luigi.ListParameter()
    docker_environment_variables: tuple[str, ...] = luigi.ListParameter()
    accelerator: tuple[str, ...] = luigi.ListParameter()



from exasol_integration_test_docker_environment.lib.base.info import Info


class RequiredTaskInfo(Info):
    def __init__(self, module_name: str, class_name: str, params: dict) -> None:
        self.params = params
        self.class_name = class_name
        self.module_name = module_name


class RequiredTaskInfoDict:
    def __init__(self, infos: dict[str, RequiredTaskInfo]) -> None:
        self.infos = infos



import click

push_options = [
    click.option(
        "--force-push/--no-force-push",
        default=False,
        help="Forces the system to overwrite existing images in registry for build steps that run",
    ),
    click.option(
        "--push-all/--no-push-all",
        default=False,
        help="Forces the system to push all images of build-steps that are specified by the goals",
    ),
]



import hashlib
import json
import logging
import shutil
from collections.abc import Generator
from pathlib import Path
from typing import (
    Any,
    TypeVar,
    Union,
)

import luigi
from luigi import (
    Task,
    util,
)
from luigi.parameter import ParameterVisibility
from luigi.task import TASK_ID_TRUNCATE_HASH

from exasol_integration_test_docker_environment.abstract_method_exception import (
    AbstractMethodException,
)
from exasol_integration_test_docker_environment.lib.base.abstract_task_future import (
    AbstractTaskFuture,
)
from exasol_integration_test_docker_environment.lib.base.pickle_target import (
    PickleTarget,
)
from exasol_integration_test_docker_environment.lib.base.task_logger_wrapper import (
    TaskLoggerWrapper,
)
from exasol_integration_test_docker_environment.lib.base.task_state import TaskState
from exasol_integration_test_docker_environment.lib.base.wrong_task_state_exception import (
    WrongTaskStateException,
)
from exasol_integration_test_docker_environment.lib.models.config.build_config import (
    build_config,
)

RETURN_TARGETS = "return_targets"

COMPLETION_TARGET = "completion_target"

RUN_DEPENDENCIES = "run_dependencies"


class RequiresTaskFuture(AbstractTaskFuture):
    """
    Represents a future for a statically ("requires") registered task.
    Details:
    - The RegisterTaskFuture is used with dependencies returned in luigis required method
    - In the required method you return tasks and usually you would to get the output target of your child tasks
      via luigis input() method (check https://luigi.readthedocs.io/en/stable/tasks.html#task-input)
    - However, if you have multiple child tasks which return in requires, for example in a dict or list,
      you need address each task in the input first and then you address the output
    - output can also return a nested dict or list of targets
    - requires can return a nested dict or list of tasks
    - we reduced the output to a single completion target with the list of return value targets
      and the child tasks are collected in the registered_tasks list
    - The RegisterTaskFuture gets the position of the task in the registered_tasks as input, such as current task
    - If the user call get_output the RegisterTaskFuture call the input() method of the current task and
       uses the position of child task to retrieve the completion target of the child task
    - It then reads the completion target to get the list of return value targets and reads the requested one.
    """

    def __init__(self, current_task: "BaseTask", child_task_index: int) -> None:
        self._child_task_index = child_task_index
        self._current_task = current_task
        self._outputs_cache = None

    def get_output(self) -> Any:
        if self._current_task._task_state == TaskState.RUN:
            if self._outputs_cache is None:
                completion_target = self._current_task.input()[self._child_task_index]
                self._outputs_cache = completion_target.read()
            return self._outputs_cache
        else:
            raise WrongTaskStateException(
                self._current_task._task_state, "RequiresTaskFuture.read_outputs_dict"
            )


class RunTaskFuture(AbstractTaskFuture):
    """
    Represents a future for a dynamically registered task.
    Details:
    - run_dependencies uses luigi dynamic task by yielding the tasks
    - luigi returns for each yielded task the return value of that tasks output method
      (see https://luigi.readthedocs.io/en/stable/tasks.html#dynamic-dependencies)
    - the output method returns targets and in case of the BaseTask this is always the completion target
      which contains the list of return value targets
    - the RunTaskFuture encapsulate the target returned by the yield and handles reading the target for the user,
      such that the user get directly the return values of the child task

    """

    def __init__(self, completion_target: PickleTarget) -> None:
        self._outputs_cache = None
        self.completion_target = completion_target

    def get_output(self) -> Any:
        if self._outputs_cache is None:
            self._outputs_cache = self.completion_target.read()
        return self._outputs_cache


# We define this generic type here, because we want to define `BaseTask.run_dependencies` to return a Generator with
# this type as yield type. This allows us to define the specific types later when using `run_dependencies` in the subclasses,
# which makes maintenance of the code much easier.
# If we would use `BaseTask` as yield type in the return type of `run_dependencies`, we would be limited to use
# `BaseTask` in the subclasses, too.
BaseTaskType = TypeVar("BaseTaskType", bound="BaseTask")


class BaseTask(Task):
    caller_output_path: tuple[str] = luigi.ListParameter(
        default=[], significant=False, visibility=ParameterVisibility.HIDDEN
    )
    job_id: str = luigi.Parameter()

    def __init__(self, *args, **kwargs) -> None:
        self._registered_tasks: list["BaseTask"] = []
        self._run_dependencies_tasks: list["BaseTask"] = []
        self._task_state = TaskState.INIT
        super().__init__(*args, **kwargs)
        self.task_id = self.task_id_str(
            self.get_task_family(), self.get_parameter_as_string_dict()
        )
        self.__hash = hash(self.task_id)
        self._init_non_pickle_attributes()
        self.register_required()
        self._task_state = TaskState.AFTER_INIT

    def _init_non_pickle_attributes(self) -> None:
        logger = logging.getLogger(f"luigi-interface.{self.__class__.__name__}")
        self.logger = TaskLoggerWrapper(logger, self.__repr__())
        self._run_dependencies_target = PickleTarget(
            path=self._get_tmp_path_for_run_dependencies()
        )
        self._complete_target = PickleTarget(
            path=self._get_tmp_path_for_completion_target()
        )
        self._registered_return_target = None

    def __getstate__(self):
        new_dict = dict(self.__dict__)
        del new_dict["logger"]
        del new_dict["_complete_target"]
        del new_dict["_run_dependencies_target"]
        return new_dict

    def __setstate__(self, new_dict):
        self.__dict__ = new_dict
        self._init_non_pickle_attributes()

    def task_id_str(self, task_family: str, params: dict[str, str]) -> str:
        """
        Returns a canonical string used to identify a particular task

        :param task_family: The task family (class name) of the task
        :param params: a dict mapping parameter names to their serialized values
        :return: A unique, shortened identifier corresponding to the family and params
        """
        # task_id is a concatenation of task family, the first values of the first 3 parameters
        # sorted by parameter name and a md5hash of the family/parameters as a cananocalised json.
        param_str = json.dumps(params, separators=(",", ":"), sort_keys=True)
        hash_input = param_str
        param_hash = hashlib.sha3_256(hash_input.encode("utf-8")).hexdigest()
        return f"{task_family}_{param_hash[:TASK_ID_TRUNCATE_HASH]}"

    def get_parameter_as_string_dict(self) -> dict[str, str]:
        """
        Convert all parameters to a str->str hash.
        """
        params_str = {}
        params = dict(self.get_params())
        for param_name, param_value in self.param_kwargs.items():
            if params[param_name].significant:
                params_str[param_name] = params[param_name].serialize(param_value)
        return params_str

    def get_output_path(self) -> Path:
        path = Path(
            self._get_output_path_for_job(),
            "outputs",
            Path(*self._extend_output_path()),
        )
        path.mkdir(parents=True, exist_ok=True)
        return path

    def _get_output_path_for_job(self) -> Path:
        return Path(build_config().output_directory, "jobs", self.job_id)

    def _extend_output_path(self) -> Union[tuple[str, ...], str]:
        extension = self.extend_output_path()
        if extension is None or extension == ():
            return self.task_id
        else:
            return extension

    def extend_output_path(self) -> tuple[str, ...]:
        return tuple(self.caller_output_path) + (self.task_id,)

    def _get_tmp_path_for_job(self) -> Path:
        return Path(self._get_output_path_for_job(), "temp")

    def _get_tmp_path_for_task(self) -> Path:
        return Path(self._get_tmp_path_for_job(), self.task_id)

    def _get_tmp_path_for_completion_target(self) -> Path:
        return Path(self._get_tmp_path_for_task(), COMPLETION_TARGET)

    def _get_tmp_path_for_run_dependencies(self) -> Path:
        return Path(self._get_tmp_path_for_task(), RUN_DEPENDENCIES)

    def get_log_path(self) -> Path:
        path = Path(self.get_output_path(), "logs")
        path.mkdir(parents=True, exist_ok=True)
        return path

    def get_cache_path(self) -> Path:
        path = Path(str(build_config().output_directory), "cache")
        path.mkdir(parents=True, exist_ok=True)
        return path

    def register_required(self) -> None:
        pass

    def register_dependency(self, task: BaseTaskType) -> RequiresTaskFuture:
        """
        Registers a 'requires' (a static) dependency of the task. It returns a future which can be used in the
        run_task() method via BaseTask.get_values_from_future() to get access to the result of the child task.
        See class RequiresTaskFuture for more details.
        """
        if self._task_state == TaskState.INIT:
            index = len(self._registered_tasks)
            self._registered_tasks.append(task)
            return RequiresTaskFuture(self, index)
        else:
            raise WrongTaskStateException(self._task_state, "register_dependency")

    def register_dependencies(self, tasks) -> Any:
        if isinstance(tasks, dict):
            return {
                key: self.register_dependencies(task) for key, task in tasks.items()
            }
        elif isinstance(tasks, list):
            return [self.register_dependencies(task) for task in tasks]
        elif isinstance(tasks, BaseTask):
            return self.register_dependency(tasks)
        else:
            return tasks

    def get_values_from_futures(self, futures) -> Any:
        if isinstance(futures, dict):
            return {
                key: self.get_values_from_futures(task) for key, task in futures.items()
            }
        elif isinstance(futures, list):
            return [self.get_values_from_futures(task) for task in futures]
        elif isinstance(futures, AbstractTaskFuture):
            return self.get_values_from_future(futures)
        else:
            return futures

    def get_values_from_future(self, future: AbstractTaskFuture) -> Any:
        return future.get_output()

    def requires(self):
        return self._registered_tasks

    def output(self):
        return self._complete_target

    def run(self):
        try:
            self._task_state = TaskState.RUN
            task_generator = self.run_task()
            if task_generator is not None:
                yield from task_generator
            self._task_state = TaskState.FINISHED
            self.logger.info("Write complete_target")
            self._complete_target.write(self._registered_return_target)
        except Exception as e:
            self._task_state = TaskState.ERROR
            self.logger.exception("Exception in run: %s", e)
            raise e

    def run_task(self):
        raise AbstractMethodException()

    def run_dependencies(self, tasks) -> Generator[BaseTaskType, Any, Any]:
        """
        Runs a 'run' (a dynamic) dependency
        (that means a dependencies which was evaluated during the runtime of the task), and returns a RunTaskFuture.
        The returned future can then be used with BaseTask.get_values_from_futures() to get access to the result
        of the child task.
        """
        if self._task_state == TaskState.RUN:
            self._register_run_dependencies(tasks)
            self._run_dependencies_target.write(self._run_dependencies_tasks)
            completion_targets = yield tasks
            task_futures = self._generate_run_task_futures(completion_targets)
            return task_futures
        else:
            raise WrongTaskStateException(self._task_state, "run_dependency")

    def _register_run_dependencies(
        self,
        tasks: Any,
    ) -> None:
        if isinstance(tasks, dict):
            for key, task in tasks.items():
                self._register_run_dependencies(task)
        elif isinstance(tasks, list):
            for task in tasks:
                self._register_run_dependencies(task)
        elif isinstance(tasks, BaseTask):
            self._run_dependencies_tasks.append(tasks)

    def _generate_run_task_futures(
        self, completion_targets: Union[Any]
    ) -> Union[list[Any], dict[Any, Any], RunTaskFuture, Any]:
        if isinstance(completion_targets, dict):
            return {
                key: self._generate_run_task_futures(task)
                for key, task in completion_targets.items()
            }
        elif isinstance(completion_targets, list):
            return [
                self._generate_run_task_futures(task) for task in completion_targets
            ]
        elif isinstance(completion_targets, PickleTarget):
            return RunTaskFuture(completion_targets)
        else:
            return completion_targets

    def return_object(self, object: Any) -> None:
        """Returns the object to the calling task. The object needs to be pickleable"""
        if self._task_state == TaskState.RUN:
            if self._registered_return_target is None:
                self._registered_return_target = object
            else:
                raise Exception(f"return target already used")
        else:
            raise WrongTaskStateException(self._task_state, "return_target")

    def get_result(self) -> Any:
        """
        Returns the return value of the task,
        this means the single value which was passed with return_object() during run_task().
        Note that it's safe to call this method from the client side, ignoring luigi's scheduler, as it uses
        persistent data to get the result.
        """
        if not self.output().exists():
            # Actual state might be unknown, because we might be called from the client side.
            raise WrongTaskStateException(TaskState.NONE, "get_result")
        return self.output().read()

    def __repr__(self) -> str:
        """
        Build a task representation like `MyTask(param1=1.5, param2='5')`
        """
        params = self.get_params()
        param_values = self.get_param_values(params, [], self.param_kwargs)

        # Build up task id
        repr_parts = []
        param_objs = dict(params)
        for param_name, param_value in param_values:
            if (
                param_objs[param_name].significant
                and param_objs[param_name].visibility == ParameterVisibility.PUBLIC
            ):
                repr_parts.append(
                    "{}={}".format(
                        param_name, param_objs[param_name].serialize(param_value)
                    )
                )

        task_str = "{}({})".format(self.task_id, ", ".join(repr_parts))

        return task_str

    def create_child_task_with_common_params(
        self, task_class: type[BaseTaskType], **kwargs
    ) -> BaseTaskType:
        params = util.common_params(self, task_class)
        params["caller_output_path"] = self._extend_output_path()
        params["job_id"] = self.job_id
        params.update(kwargs)
        return task_class(**params)

    def create_child_task(
        self, task_class: type[BaseTaskType], **kwargs
    ) -> BaseTaskType:
        params: dict[str, Any] = {}
        params["caller_output_path"] = self._extend_output_path()
        params["job_id"] = self.job_id
        params.update(kwargs)
        return task_class(**params)

    def cleanup(self, success: bool) -> None:
        self.cleanup_internal(success, set())

    def cleanup_internal(self, success: bool, cleanup_checklist: set[str]) -> None:
        self.logger.debug("Cleaning up")
        if str(self) not in cleanup_checklist:
            cleanup_checklist.add(str(self))
            if (
                self._task_state != TaskState.CLEANUP
                and self._task_state != TaskState.CLEANED
            ):
                self._task_state = TaskState.CLEANUP
                try:
                    self.cleanup_child_task(success, cleanup_checklist)
                except Exception as e:
                    self.logger.error("Error during cleaning up child tasks", e)
                try:
                    self.cleanup_task(success)
                except Exception as e:
                    self.logger.error("Error during cleaning up task", e)
                if self._get_tmp_path_for_task().exists():
                    shutil.rmtree(self._get_tmp_path_for_task())
                self._task_state = TaskState.CLEANED
            self.logger.debug("Cleanup finished")
        else:
            self.logger.debug("Cleanup skipped")

    def cleanup_child_task(self, success: bool, cleanup_checklist: set[str]) -> None:
        if self._run_dependencies_target.exists():
            _run_dependencies_tasks_from_target = self._run_dependencies_target.read()
        else:
            _run_dependencies_tasks_from_target = []
        _run_dependencies_tasks = (
            self._run_dependencies_tasks + _run_dependencies_tasks_from_target
        )
        reversed_run_dependencies_task_list = list(_run_dependencies_tasks)
        reversed_run_dependencies_task_list.reverse()
        for task in reversed_run_dependencies_task_list:
            task.cleanup_internal(success, cleanup_checklist)

        reversed_registered_task_list = list(self._registered_tasks)
        reversed_registered_task_list.reverse()
        for task in reversed_registered_task_list:
            task.cleanup_internal(success, cleanup_checklist)

    def cleanup_task(self, success: bool) -> None:
        pass



from exasol_integration_test_docker_environment.lib.api.build_test_container import (
    build_test_container,
)
from exasol_integration_test_docker_environment.lib.api.health import health
from exasol_integration_test_docker_environment.lib.api.push_test_container import (
    push_test_container,
)
from exasol_integration_test_docker_environment.lib.api.spawn_test_environment import (
    spawn_test_environment,
)
from exasol_integration_test_docker_environment.lib.api.spawn_test_environment_with_test_container import (
    spawn_test_environment_with_test_container,
)



from exasol_integration_test_docker_environment.lib.base.info import Info


class DockerVolumeInfo(Info):

    def __str__(self):
        return (
            f"DockerVolumeInfo: \n"
            f'volume_name="{self.volume_name}"\n'
            f"reused={self.reused}\n"
            f"mount_point={self.mount_point}"
        )

    def __init__(self, volume_name: str, mount_point: str, reused: bool = False):
        self.volume_name = volume_name
        self.reused = reused
        self.mount_point = mount_point



import click

from exasol_integration_test_docker_environment.cli.options.system_options import (
    output_directory_option,
    tempory_base_directory_option,
)

build_options = [
    click.option(
        "--force-rebuild/--no-force-rebuild",
        default=False,
        help="Forces the system to complete rebuild all stages down to the stages "
        "specified with the options --force-rebuild-from.",
    ),
    click.option(
        "--force-rebuild-from",
        multiple=True,
        type=str,
        help="If the option --force-rebuild is given, "
        "this options specifies for which stages and dependent stages system will force a rebuild. "
        "The option can be repeated with different stages. "
        "The system will than force the rebuild of these stages and their. dependet stages.",
    ),
    click.option(
        "--force-pull/--no-force-pull",
        default=False,
        help="Forces the system to pull all stages if available, otherwise it rebuilds a stage.",
    ),
    output_directory_option,
    tempory_base_directory_option,
    click.option(
        "--log-build-context-content/--no-log-build-context-content",
        default=False,
        help="For Debugging: Logs the files and directories in the build context of a stage",
    ),
    click.option(
        "--cache-directory",
        default=None,
        type=click.Path(file_okay=False, dir_okay=True, exists=False),
        help="Directory from where saved docker images can be loaded",
    ),
    click.option(
        "--build-name",
        default=None,
        type=str,
        help="Name of the build. For example: Repository + CI Build Number",
    ),
]



import math
import time
from pathlib import Path
from threading import Thread
from typing import (
    Callable,
    Optional,
)

from docker.models.containers import Container

from exasol_integration_test_docker_environment.lib.base.still_running_logger import (
    StillRunningLogger,
)
from exasol_integration_test_docker_environment.lib.base.task_logger_wrapper import (
    TaskLoggerWrapper,
)
from exasol_integration_test_docker_environment.lib.logging.container_log_handler import (
    ContainerLogHandler,
)


class DBContainerLogThread(Thread):
    def __init__(
        self,
        container: Container,
        logger: TaskLoggerWrapper,
        log_file: Path,
        description: str,
    ) -> None:
        super().__init__()
        self.complete_log: list[str] = []
        self.description = description
        self.logger = logger
        self.log_file = log_file
        self.container = container
        self.finish = False
        self.previous_timestamp: Optional[float] = None
        self.current_timestamp: Optional[float] = None
        self.error_message: Optional[str] = None
        self.ignore_error_return_codes = (
            "(membership) returned with state 1",  # exclude webui not found in 7.0.0
            "rsyslogd) returned with state 1",  # exclude rsyslogd which might crash when running itde under lima
        )

    def _contains_error(self, log_line: str) -> bool:
        def ignore_sshd(log_line_local):
            return "sshd was not started" in log_line_local

        def ignore_return_code(log_line_local):
            return any(x in log_line_local for x in self.ignore_error_return_codes)

        def contains(substr: str, ignore: Optional[Callable[[str], bool]] = None):
            if not substr in log_line:
                return False
            return ignore is None or not ignore(log_line)

        return (
            contains("error", ignore_sshd)
            or contains("exception")
            or contains("returned with state 1", ignore_return_code)
        )

    def stop(self):
        self.logger.info("Stop ContainerLogThread")
        self.finish = True

    def run(self):
        try:
            with ContainerLogHandler(
                self.log_file, self.logger, self.description
            ) as log_handler:
                still_running_logger = StillRunningLogger(self.logger, self.description)
                while not self.finish:
                    self.current_timestamp = math.floor(time.time())
                    log = self.container.logs(
                        since=self.previous_timestamp, until=self.current_timestamp
                    )
                    if len(log) != 0:
                        still_running_logger.log()
                        log_handler.handle_log_lines(log)
                    log_line = log.decode("utf-8").lower()
                    if self._contains_error(log_line):
                        self.logger.info(
                            "ContainerLogHandler error message, %s", log_line
                        )
                        self.error_message = log_line
                        self.finish = True
                    self.previous_timestamp = self.current_timestamp
                    self.complete_log = log_handler.get_complete_log().copy()
                    time.sleep(1)
        except Exception as e:
            self.finish = True
            self.logger.exception("Caught exception in DBContainerLogThread.run.")



import luigi

from exasol_integration_test_docker_environment.lib.base.dependency_logger_base_task import (
    DependencyLoggerBaseTask,
)
from exasol_integration_test_docker_environment.lib.models.data.environment_type import (
    EnvironmentType,
)
from exasol_integration_test_docker_environment.lib.test_environment.parameter.spawn_test_environment_parameter import (
    SpawnTestEnvironmentParameter,
)
from exasol_integration_test_docker_environment.lib.test_environment.spawn_test_environment_with_docker_db import (
    SpawnTestEnvironmentWithDockerDB,
)
from exasol_integration_test_docker_environment.lib.test_environment.spawn_test_environment_with_external_db import (
    SpawnTestEnvironmentWithExternalDB,
)


class SpawnTestEnvironment(DependencyLoggerBaseTask, SpawnTestEnvironmentParameter):
    DEFAULT_DB_USER = "sys"
    DEFAULT_DATABASE_PASSWORD = "exasol"
    DEFAULT_BUCKETFS_WRITE_PASSWORD = "write"
    environment_name: str = luigi.Parameter()

    def register_required(self):
        task = self._create_spawn_environment_task()
        self._environment_info_future = self.register_dependency(task)

    def _create_spawn_environment_task(self):
        if self.environment_type == EnvironmentType.docker_db:
            return self._create_docker_db_environment()
        else:
            return self._create_external_db_environment()

    def _create_external_db_environment(self):
        if self.external_exasol_db_host is None:
            raise Exception("external_exasol_db_host not set")
        if self.external_exasol_db_port is None:
            raise Exception("external_exasol_db_port not set")
        if self.external_exasol_bucketfs_http_port is None:
            raise Exception("external_exasol_bucketfs_http_port not set")
        task = self.create_child_task_with_common_params(
            SpawnTestEnvironmentWithExternalDB,
            db_user=self.external_exasol_db_user,
            db_password=self.external_exasol_db_password,
            bucketfs_write_password=self.external_exasol_bucketfs_write_password,
        )
        return task

    def _create_docker_db_environment(self):
        task = self.create_child_task_with_common_params(
            SpawnTestEnvironmentWithDockerDB,
            db_user=self.DEFAULT_DB_USER,
            db_password=self.DEFAULT_DATABASE_PASSWORD,
            bucketfs_write_password=self.DEFAULT_BUCKETFS_WRITE_PASSWORD,
        )
        return task

    def run_task(self) -> None:
        environment_info = self.get_values_from_future(self._environment_info_future)
        self.return_object(environment_info)



from typing import (
    Optional,
)

from exasol_integration_test_docker_environment.lib.models.data.environment_info import (
    EnvironmentInfo,
)


class ShellVariables:
    """
    Represents a collection of unix shell environment variables.
    """

    def __init__(self, env: dict[str, str]) -> None:
        self.env = env

    @classmethod
    def from_test_environment_info(
        cls,
        default_bridge_ip_address: Optional[str],
        test_environment_info: EnvironmentInfo,
    ) -> "ShellVariables":
        """
        Create ShellVariables from the given test_container_name,
        default_bridge_ip_address and EnvironmentInfo.
        """
        info = test_environment_info
        assert info.database_info.ports.database is not None
        assert info.database_info.ports.bucketfs_http is not None
        env: dict[str, str] = {
            "NAME": info.name,
            "TYPE": info.type if isinstance(info.type, str) else info.type.name,
            "DATABASE_HOST": info.database_info.host,
            "DATABASE_DB_PORT": str(info.database_info.ports.database),
            "DATABASE_BUCKETFS_PORT": str(info.database_info.ports.bucketfs_http),
            "DATABASE_BUCKETFS_HTTP_PORT": str(info.database_info.ports.bucketfs_http),
            "DATABASE_BUCKETFS_HTTPS_PORT": str(
                info.database_info.ports.bucketfs_https
            ),
            "DATABASE_SSH_PORT": (
                str(info.database_info.ports.ssh)
                if info.database_info.ports.ssh is not None
                else ""
            ),
        }

        if info.database_info.container_info is not None:
            assert info.database_info.container_info.volume_name
            assert default_bridge_ip_address
            network_aliases = " ".join(
                info.database_info.container_info.network_aliases
            )
            env.update(
                {
                    "DATABASE_CONTAINER_NAME": info.database_info.container_info.container_name,
                    "DATABASE_CONTAINER_NETWORK_ALIASES": f'"{network_aliases}"',
                    "DATABASE_CONTAINER_IP_ADDRESS": info.database_info.container_info.ip_address,
                    "DATABASE_CONTAINER_VOLUMNE_NAME": info.database_info.container_info.volume_name,
                    "DATABASE_CONTAINER_DEFAULT_BRIDGE_IP_ADDRESS": default_bridge_ip_address,
                }
            )
        if info.test_container_info is not None:
            container_name = info.test_container_info.container_name
            network_aliases = " ".join(info.test_container_info.network_aliases)
            env.update(
                {
                    "TEST_CONTAINER_NAME": container_name,
                    "TEST_CONTAINER_NETWORK_ALIASES": f'"{network_aliases}"',
                    "TEST_CONTAINER_IP_ADDRESS": info.test_container_info.ip_address,
                }
            )
        return ShellVariables(env)

    def render(self, prefix: str = "") -> str:
        prefix += "ITDE"
        aslist = [f"{prefix}_{key}={value}" for key, value in self.env.items()]
        return "\n".join(aslist) + "\n"



from inspect import cleandoc

from exasol_integration_test_docker_environment.doctor import health_checkup
from exasol_integration_test_docker_environment.lib.models.api_errors import (
    HealthProblem,
)
from exasol_integration_test_docker_environment.lib.utils.api_function_decorators import (
    cli_function,
)


@cli_function
def health():
    """
    Check the health of the execution environment.

    If no issues have been found, using the library or executing the test should work just fine.
    For all found issues there will be a proposed fix/solution.

    If the environment was found to be healthy it will return None, otherwise a description of the found issues.
    :raises HealthProblem
    """

    errors = set(health_checkup())
    if not errors:
        return

    message = cleandoc(
        """
        {count} problem(s) have been identified.

        {problems}
        """
    ).format(
        count=len(errors),
        problems="\n".join(f"{error.value}" for error in errors),
    )
    raise HealthProblem(message)



import gzip
import pathlib
import time
from datetime import (
    datetime,
    timedelta,
)

import luigi
from docker.models.containers import Container

from exasol_integration_test_docker_environment.lib.base.db_os_executor import (
    DbOsExecFactory,
)
from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import (
    JsonPickleParameter,
)
from exasol_integration_test_docker_environment.lib.models.data.database_credentials import (
    DatabaseCredentialsParameter,
)
from exasol_integration_test_docker_environment.lib.models.data.database_info import (
    DatabaseInfo,
)
from exasol_integration_test_docker_environment.lib.test_environment.database_waiters.db_container_log_thread import (
    DBContainerLogThread,
)
from exasol_integration_test_docker_environment.lib.test_environment.database_waiters.is_database_ready_thread import (
    IsDatabaseReadyThread,
)


class WaitForTestDockerDatabase(DockerBaseTask, DatabaseCredentialsParameter):
    environment_name: str = luigi.Parameter()
    database_info: DatabaseInfo = JsonPickleParameter(DatabaseInfo, significant=False)  # type: ignore
    db_startup_timeout_in_seconds: int = luigi.IntParameter(
        default=10 * 60, significant=False
    )
    attempt: int = luigi.IntParameter(default=1)
    docker_db_image_version: str = luigi.Parameter()
    executor_factory: DbOsExecFactory = JsonPickleParameter(DbOsExecFactory, significant=False)  # type: ignore

    def run_task(self) -> None:
        with self._get_docker_client() as docker_client:
            assert self.database_info.container_info is not None
            db_container_name = self.database_info.container_info.container_name
            db_container = docker_client.containers.get(db_container_name)
            is_database_ready = self.wait_for_database_startup(db_container)
            after_startup_db_log_file = self.get_log_path().joinpath(
                "after_startup_db_log.tar.gz"
            )
            self.save_db_log_files_as_gzip_tar(after_startup_db_log_file, db_container)
            self.return_object(is_database_ready)

    def wait_for_database_startup(self, db_container: Container) -> bool:
        container_log_thread, is_database_ready_thread = self.start_wait_threads(
            db_container
        )
        is_database_ready = self.wait_for_threads(
            container_log_thread, is_database_ready_thread
        )
        self.join_threads(container_log_thread, is_database_ready_thread)
        return is_database_ready

    def start_wait_threads(
        self, db_container: Container
    ) -> tuple[DBContainerLogThread, IsDatabaseReadyThread]:
        startup_log_file = self.get_log_path().joinpath("startup.log")
        container_log_thread = DBContainerLogThread(
            db_container,
            self.logger,
            startup_log_file,
            "Database Startup %s" % db_container.name,
        )
        container_log_thread.start()
        is_database_ready_thread = IsDatabaseReadyThread(
            self.logger,
            self.database_info,
            db_container,
            self.get_database_credentials(),
            self.docker_db_image_version,
            self.executor_factory,
        )
        is_database_ready_thread.start()
        return container_log_thread, is_database_ready_thread

    def join_threads(
        self,
        container_log_thread: DBContainerLogThread,
        is_database_ready_thread: IsDatabaseReadyThread,
    ) -> None:
        container_log_thread.stop()
        is_database_ready_thread.stop()
        container_log_thread.join()
        is_database_ready_thread.join()

    def wait_for_threads(
        self,
        container_log_thread: DBContainerLogThread,
        is_database_ready_thread: IsDatabaseReadyThread,
    ) -> bool:
        is_database_ready = False
        reason = None
        start_time = datetime.now()
        while True:
            if container_log_thread.error_message != None:
                is_database_ready = False
                reason = "error message in container log"
                break
            if is_database_ready_thread.finish and is_database_ready_thread.is_ready:
                is_database_ready = True
                break
            elif is_database_ready_thread.finish:
                reason = "error in is_database_ready_thread"
                break
            if self.timeout_occured(start_time):
                reason = (
                    f"timeout after after {self.db_startup_timeout_in_seconds} seconds"
                )
                is_database_ready = False
                break
            time.sleep(1)
        if not is_database_ready:
            self.log_database_not_ready(
                container_log_thread, is_database_ready_thread, reason
            )
        container_log_thread.stop()
        is_database_ready_thread.stop()
        return is_database_ready

    def log_database_not_ready(
        self,
        container_log_thread: DBContainerLogThread,
        is_database_ready_thread: IsDatabaseReadyThread,
        reason,
    ) -> None:
        container_log = "\n".join(container_log_thread.complete_log)
        log_information = f"""
========== IsDatabaseReadyThread output db connection: ============
{is_database_ready_thread.output_db_connection}
========== IsDatabaseReadyThread output bucketfs connection: ============
{is_database_ready_thread.output_bucketfs_connection}
========== Container-Log: ============
{container_log}
"""
        self.logger.warning(
            'Database startup failed for following reason "%s", here some debug information \n%s',
            reason,
            log_information,
        )

    def timeout_occured(self, start_time: datetime) -> bool:
        timeout = timedelta(seconds=self.db_startup_timeout_in_seconds)
        return datetime.now() - start_time > timeout

    def save_db_log_files_as_gzip_tar(
        self, path: pathlib.Path, database_container: Container
    ) -> None:
        stream, stat = database_container.get_archive("/exa/logs")
        with gzip.open(path, "wb") as file:
            for chunk in stream:
                file.write(chunk)

    def write_output(self, is_database_ready: bool) -> None:
        with self.output().open("w") as file:
            file.write(str(is_database_ready))



from typing import (
    Optional,
)

from exasol_integration_test_docker_environment.cli.options.docker_repository_options import (
    DEFAULT_DOCKER_REPOSITORY_NAME,
)
from exasol_integration_test_docker_environment.cli.options.system_options import (
    DEFAULT_OUTPUT_DIRECTORY,
)
from exasol_integration_test_docker_environment.lib.base.run_task import (
    generate_root_task,
    run_task,
)
from exasol_integration_test_docker_environment.lib.docker.images.image_info import (
    ImageInfo,
)
from exasol_integration_test_docker_environment.lib.models.config.build_config import (
    set_build_config,
)
from exasol_integration_test_docker_environment.lib.models.config.docker_config import (
    set_docker_repository_config,
)
from exasol_integration_test_docker_environment.lib.models.data.test_container_content_description import (
    TestContainerContentDescription,
)
from exasol_integration_test_docker_environment.lib.test_environment.analyze_test_container import (
    AnalyzeTestContainer,
    DockerTestContainerPush,
)
from exasol_integration_test_docker_environment.lib.utils.api_function_decorators import (
    no_cli_function,
)


@no_cli_function
def push_test_container(
    test_container_content: TestContainerContentDescription,
    force_push: bool = False,
    push_all: bool = False,
    force_rebuild: bool = False,
    force_rebuild_from: tuple[str, ...] = (),
    force_pull: bool = False,
    output_directory: str = DEFAULT_OUTPUT_DIRECTORY,
    temporary_base_directory: str = "/tmp",
    log_build_context_content: bool = False,
    cache_directory: Optional[str] = None,
    build_name: Optional[str] = None,
    source_docker_repository_name: str = DEFAULT_DOCKER_REPOSITORY_NAME,
    source_docker_tag_prefix: str = "",
    source_docker_username: Optional[str] = None,
    source_docker_password: Optional[str] = None,
    target_docker_repository_name: str = DEFAULT_DOCKER_REPOSITORY_NAME,
    target_docker_tag_prefix: str = "",
    target_docker_username: Optional[str] = None,
    target_docker_password: Optional[str] = None,
    workers: int = 5,
    task_dependencies_dot_file: Optional[str] = None,
    log_level: Optional[str] = None,
    use_job_specific_log_file: bool = False,
) -> ImageInfo:
    """
    This function pushes all stages of the test container for the test environment.
    If the stages do not exist locally, the system will build or pull them before the push.
    """
    set_build_config(
        force_rebuild,
        force_rebuild_from,
        force_pull,
        log_build_context_content,
        output_directory,
        temporary_base_directory,
        cache_directory,
        build_name,
    )
    # Use AnalyzeTestContainer to ensure that all luigi processes got it loaded
    analyze_task = AnalyzeTestContainer.__class__.__name__

    set_docker_repository_config(
        source_docker_password,
        source_docker_repository_name,
        source_docker_username,
        source_docker_tag_prefix,
        "source",
    )
    set_docker_repository_config(
        target_docker_password,
        target_docker_repository_name,
        target_docker_username,
        target_docker_tag_prefix,
        "target",
    )
    task_creator = lambda: generate_root_task(
        task_class=DockerTestContainerPush,
        test_container_content=test_container_content,
        force_push=force_push,
        push_all=push_all,
    )
    image_infos = run_task(
        task_creator,
        workers,
        task_dependencies_dot_file,
        log_level=log_level,
        use_job_specific_log_file=use_job_specific_log_file,
    )
    return image_infos[0]



import logging
import tempfile
from pathlib import Path
from types import ModuleType
from typing import Optional

import importlib_resources as ir

LOG = logging.getLogger("resource_directory")


def _copy_importlib_resources_file(
    src_file: ir.abc.Traversable, target_file: Path
) -> None:
    """
    Uses a given source path "src_file" given as an importlib_resources.abc.Traversable to copy the file it points to
    into the destination denoted by target_path.
    :param src_file: Location of the file to be copied, given as importlib_resources.abc.Traversable.
    :param target_file: Path object the location file should be copied to.
    :raises RuntimeError if parameter target_file already exists.
    """
    if src_file.name == "__init__.py":
        LOG.debug(f"Ignoring {src_file} for repository.")
        return
    if target_file.exists():
        error = f"Repository target: {target_file} already exists."
        LOG.error(error)
        raise RuntimeError(error)

    content = src_file.read_bytes()
    with open(target_file, "wb") as file:
        file.write(content)


def _copy_importlib_resources_dir_tree(
    src_path: ir.abc.Traversable, target_path: Path
) -> None:
    """
    Uses a given source path "scr_path" given as an importlib_resources.abc.Traversable to copy all files/directories
    in the directory tree whose root is scr_path into target_path.
    :param src_path: Root of the dir tree to be copied, given as importlib_resources.abc.Traversable.
    :param target_path: Path object the dir tree should be copied to.
    :raises RuntimeError if parameter target_file already exists.
    """
    if not target_path.exists():
        target_path.mkdir()
    for file in src_path.iterdir():
        file_target = target_path / file.name
        if file.is_file():
            _copy_importlib_resources_file(file, file_target)
        else:
            file_target.mkdir(exist_ok=True)
            _copy_importlib_resources_dir_tree(file, file_target)


class ResourceDirectory:
    """
    Copies all files stored within the given package (resource_package) into a temporary directory
    and yields the directory name.
    :resource_package The module whose content should be copied to a temporary directory.
    """

    def __init__(self, resource_package: ModuleType) -> None:
        # We need to transform the module to a string and later back to a module
        # because this class will be pickled by luigi and modules are not supported for serialization
        self._resource_package_str = resource_package.__name__
        self._tmp_directory: Optional[tempfile.TemporaryDirectory] = None

    @property
    def tmp_directory(self):
        if self._tmp_directory is not None:
            return self._tmp_directory.name
        else:
            return None

    def create(self) -> str:
        self._tmp_directory = tempfile.TemporaryDirectory()
        assert self._tmp_directory
        source_path = ir.files(self._resource_package_str)
        LOG.debug(
            f"Copying resource package: '{self._resource_package_str}' to '{self._tmp_directory.name}'"
        )
        _copy_importlib_resources_dir_tree(source_path, Path(self._tmp_directory.name))
        return self._tmp_directory.name

    def cleanup(self):
        if self._tmp_directory is not None:
            self._tmp_directory.cleanup()
            self._tmp_directory = None

    def __enter__(self):
        return self.create()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()



from luigi.parameter import ParameterVisibility

from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import (
    JsonPickleParameter,
)
from exasol_integration_test_docker_environment.lib.models.data.test_container_content_description import (
    TestContainerContentDescription,
)


class TestContainerParameter:
    test_container_content: TestContainerContentDescription = JsonPickleParameter(
        TestContainerContentDescription, visibility=ParameterVisibility.HIDDEN
    )  # type: ignore


class OptionalTestContainerParameter:
    test_container_content: TestContainerContentDescription = JsonPickleParameter(
        TestContainerContentDescription,
        visibility=ParameterVisibility.HIDDEN,
        is_optional=True,
    )  # type: ignore



from enum import Enum

import jsonpickle


class TaskDescription:

    def __init__(self, id: str, representation: str) -> None:
        self.representation = representation
        self.id = id

    @property
    def formatted_representation(self):
        representation = self.representation.replace('"', '\\"')
        return f'"{representation}"'

    @property
    def formatted_id(self):
        assert '"' not in self.id
        return f'"{self.id}"'


class DependencyType(Enum):
    requires = (1,)
    dynamic = 2


class DependencyState(Enum):
    requested = 1
    finished = 2


class TaskDependency:

    def __init__(
        self,
        source: TaskDescription,
        target: TaskDescription,
        type: DependencyType,
        index: int,
        state: DependencyState,
    ) -> None:
        self.state = state.name
        self.index = index
        self.type = type.name
        self.target = target
        self.source = source

    def to_json(self):
        jsonpickle.set_preferred_backend("simplejson")
        jsonpickle.set_encoder_options("simplejson", sort_keys=True)
        return jsonpickle.encode(self)

    @classmethod
    def from_json(cls, json_string: object) -> object:
        loaded_object = jsonpickle.decode(json_string)
        if not isinstance(loaded_object, cls):
            raise TypeError(
                "Type {} of loaded object does not match {}".format(
                    type(loaded_object), cls
                )
            )
        return loaded_object

    @property
    def formatted(self):
        assert '"' not in str(self)
        return f'"{str(self)}"'

    def __str__(self):
        return (
            f"TaskDependency(source={self.source}, "
            f"target={self.target}, type={self.type}, "
            f"index={self.index}, state={self.state})"
        )



from typing import (
    Optional,
)

from exasol_integration_test_docker_environment.lib.base.info import Info
from exasol_integration_test_docker_environment.lib.models.data.docker_network_info import (
    DockerNetworkInfo,
)


class ContainerInfo(Info):

    def __init__(
        self,
        container_name: str,
        ip_address: str,
        network_aliases: list[str],
        network_info: DockerNetworkInfo,
        volume_name: Optional[str] = None,
    ) -> None:
        self.network_aliases = network_aliases
        self.ip_address = ip_address
        self.network_info = network_info
        self.container_name = container_name
        self.volume_name = volume_name



from collections.abc import Iterable
from typing import (
    Optional,
)


class ArgumentConstraintError(ValueError):
    """Represents an error for an argument not fulfilling a specific constraint"""


class HealthProblem(RuntimeError):
    """Represents a problem found by analyzing the project health"""


class TaskFailures(Exception):
    """Represents a potential cause of a TaskRuntimeError"""

    def __init__(self, inner: Optional[list[str]] = None):
        super().__init__(self._construct_exception_message(inner))
        self.inner = inner

    def _construct_exception_message(self, failures: Optional[Iterable[str]]) -> str:
        if failures is not None:
            formatted_task_failures = "\n".join(failures)
            return f"Following task failures were caught during the execution:\n{formatted_task_failures}"
        else:
            return f"No task failures were caught during the execution:"


class TaskRuntimeError(RuntimeError):
    """Represents an error which occurred during execution of a luigi task"""

    def __init__(self, msg: str, inner: Optional[list[str]] = None):
        """
        Creates a TaskRuntimeError
        Args:
            msg: The error message
            inner: A list of task failures that caused this exception
                   (@deprecated will be replaced by cause TaskFailures)
        """
        super().__init__(msg)
        self.msg = msg
        self.inner = inner



from collections.abc import (
    Generator,
    Iterator,
)
from pathlib import Path
from typing import (
    Optional,
    Union,
)

import luigi

from exasol_integration_test_docker_environment.abstract_method_exception import (
    AbstractMethodException,
)
from exasol_integration_test_docker_environment.lib.base.base_task import (
    BaseTask,
    BaseTaskType,
)
from exasol_integration_test_docker_environment.lib.base.docker_base_task import (
    DockerBaseTask,
)
from exasol_integration_test_docker_environment.lib.docker.container.utils import (
    default_bridge_ip_address,
)
from exasol_integration_test_docker_environment.lib.models.data.container_info import (
    ContainerInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.database_credentials import (
    DatabaseCredentialsParameter,
)
from exasol_integration_test_docker_environment.lib.models.data.database_info import (
    DatabaseInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.docker_network_info import (
    DockerNetworkInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.docker_volume_info import (
    DockerVolumeInfo,
)
from exasol_integration_test_docker_environment.lib.models.data.environment_info import (
    EnvironmentInfo,
)
from exasol_integration_test_docker_environment.lib.test_environment.create_certificates.create_ssl_certificates_task import (
    CreateSSLCertificatesTask,
)
from exasol_integration_test_docker_environment.lib.test_environment.database_waiters.wait_for_external_database import (
    WaitForTestExternalDatabase,
)
from exasol_integration_test_docker_environment.lib.test_environment.database_waiters.wait_for_test_docker_database import (
    WaitForTestDockerDatabase,
)
from exasol_integration_test_docker_environment.lib.test_environment.docker_container_copy import (
    DockerContainerCopy,
)
from exasol_integration_test_docker_environment.lib.test_environment.parameter.general_spawn_test_environment_parameter import (
    GeneralSpawnTestEnvironmentParameter,
)
from exasol_integration_test_docker_environment.lib.test_environment.prepare_network_for_test_environment import (
    PrepareDockerNetworkForTestEnvironment,
)
from exasol_integration_test_docker_environment.lib.test_environment.shell_variables import (
    ShellVariables,
)
from exasol_integration_test_docker_environment.lib.test_environment.spawn_test_container import (
    SpawnTestContainer,
)

DATABASE = "database"

TEST_CONTAINER = "test_container"


class AbstractSpawnTestEnvironment(
    DockerBaseTask, GeneralSpawnTestEnvironmentParameter, DatabaseCredentialsParameter
):
    environment_name: str = luigi.Parameter()

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.test_container_name = f"""test_container_{self.environment_name}"""
        self.network_name = f"""db_network_{self.environment_name}"""

    def get_environment_type(self):
        raise AbstractMethodException()

    def run_task(
        self,
    ) -> Iterator[
        Union[
            PrepareDockerNetworkForTestEnvironment,
            CreateSSLCertificatesTask,
            WaitForTestDockerDatabase,
            WaitForTestExternalDatabase,
        ]
    ]:
        test_environment_info = yield from self._attempt_database_start()
        self.return_object(test_environment_info)

    def _attempt_database_start(
        self,
    ) -> Generator[
        Union[
            PrepareDockerNetworkForTestEnvironment,
            CreateSSLCertificatesTask,
            WaitForTestDockerDatabase,
            WaitForTestExternalDatabase,
        ],
        None,
        EnvironmentInfo,
    ]:
        is_database_ready = False
        attempt = 0
        database_info = None
        test_container_info = None
        while not is_database_ready and attempt < self.max_start_attempts:
            network_info, database_info, is_database_ready, test_container_info = (
                yield from self._start_database(attempt)
            )
            attempt += 1
        if not is_database_ready and not attempt < self.max_start_attempts:
            raise Exception(
                f"Maximum attempts {attempt} to start the database reached."
            )
        assert database_info is not None
        test_environment_info = EnvironmentInfo(
            name=self.environment_name,
            env_type=self.get_environment_type(),
            database_info=database_info,
            test_container_info=test_container_info,
            network_info=network_info,
        )
        self.create_test_environment_info_in_test_container_and_on_host(
            test_environment_info
        )
        return test_environment_info

    def create_test_environment_info_in_test_container(
        self,
        test_environment_info: EnvironmentInfo,
        shell_variables: ShellVariables,
        json: str,
    ):
        assert test_environment_info.test_container_info
        test_container_name = test_environment_info.test_container_info.container_name
        with self._get_docker_client() as docker_client:
            test_container = docker_client.containers.get(test_container_name)
            self.logger.info(
                f"Create test environment info in test container '{test_container_name}' at '/'"
            )
            copy = DockerContainerCopy(test_container)
            copy.add_string_to_file("environment_info.json", json)
            copy.add_string_to_file("environment_info.conf", shell_variables.render())
            copy.add_string_to_file(
                "environment_info.sh", shell_variables.render("export ")
            )
            copy.copy("/")

    def create_test_environment_info_in_test_container_and_on_host(
        self, test_environment_info: EnvironmentInfo
    ):
        path = Path(self.get_cache_path(), f"environments/{self.environment_name}")
        path.mkdir(exist_ok=True, parents=True)
        self.logger.info(f"Create test environment info on the host at '{path}'")

        json = test_environment_info.to_json()
        with Path(path, "environment_info.json").open("w") as f:
            f.write(json)

        shell_variables = self.collect_shell_variables(test_environment_info)
        with Path(path, "environment_info.conf").open("w") as f:
            f.write(shell_variables.render())

        with Path(path, "environment_info.sh").open("w") as f:
            f.write(shell_variables.render("export "))

        if test_environment_info.test_container_info is not None:
            self.create_test_environment_info_in_test_container(
                test_environment_info,
                shell_variables,
                json,
            )

    def _default_bridge_ip_address(self, test_environment_info) -> Optional[str]:
        if test_environment_info.database_info.container_info is not None:
            container_name = (
                test_environment_info.database_info.container_info.container_name
            )
            with self._get_docker_client() as docker_client:
                db_container = docker_client.containers.get(container_name)
                return default_bridge_ip_address(db_container)
        return None

    def collect_shell_variables(self, test_environment_info) -> ShellVariables:
        return ShellVariables.from_test_environment_info(
            self._default_bridge_ip_address(test_environment_info),
            test_environment_info,
        )

    def _start_database(self, attempt) -> Generator[
        Union[
            PrepareDockerNetworkForTestEnvironment,
            CreateSSLCertificatesTask,
            WaitForTestDockerDatabase,
            WaitForTestExternalDatabase,
        ],
        None,
        tuple[DockerNetworkInfo, DatabaseInfo, bool, Optional[ContainerInfo]],
    ]:
        network_info = yield from self._create_network(attempt)
        ssl_volume_info = None
        if self.create_certificates:
            ssl_volume_info = yield from self._create_ssl_certificates()
        database_info, test_container_info = (
            yield from self._spawn_database_and_test_container(
                network_info, ssl_volume_info, attempt
            )
        )
        is_database_ready = yield from self._wait_for_database(database_info, attempt)
        return network_info, database_info, is_database_ready, test_container_info

    def _create_ssl_certificates(
        self,
    ) -> Generator[CreateSSLCertificatesTask, None, Optional[DockerVolumeInfo]]:
        ssl_volume_info_future = yield from self.run_dependencies(
            self.create_ssl_certificates()
        )
        ssl_volume_info: Optional[DockerVolumeInfo] = self.get_values_from_future(
            ssl_volume_info_future
        )
        return ssl_volume_info

    def create_ssl_certificates(self):
        raise AbstractMethodException()

    def _create_network(
        self, attempt
    ) -> Generator[PrepareDockerNetworkForTestEnvironment, None, DockerNetworkInfo]:
        network_info_future = yield from self.run_dependencies(
            self.create_network_task(attempt)
        )
        network_info = self.get_values_from_future(network_info_future)
        return network_info

    def create_network_task(
        self, attempt: int
    ) -> Iterator[PrepareDockerNetworkForTestEnvironment]:
        raise AbstractMethodException()

    def _spawn_database_and_test_container(
        self,
        network_info: DockerNetworkInfo,
        certificate_volume_info: Optional[DockerVolumeInfo],
        attempt: int,
    ) -> Generator[BaseTaskType, None, tuple[DatabaseInfo, Optional[ContainerInfo]]]:
        def volume_name(info):
            return None if info is None else info.volume_name

        child_tasks = {
            DATABASE: self.create_spawn_database_task(
                network_info,
                certificate_volume_info,
                attempt,
            )
        }
        if self.test_container_content is not None:
            certificate_volume_name = volume_name(certificate_volume_info)
            child_tasks[TEST_CONTAINER] = self.create_spawn_test_container_task(
                network_info,
                certificate_volume_name,
                attempt,
            )
        futures = yield from self.run_dependencies(child_tasks)
        results = self.get_values_from_futures(futures)
        database_info: DatabaseInfo = results[DATABASE]  # type: ignore
        test_container_info: Optional[ContainerInfo] = results[TEST_CONTAINER] if self.test_container_content is not None else None  # type: ignore
        return database_info, test_container_info

    def create_spawn_database_task(
        self,
        network_info: DockerNetworkInfo,
        certificate_volume_info: Optional[DockerVolumeInfo],
        attempt: int,
    ):
        raise AbstractMethodException()

    def create_spawn_test_container_task(
        self,
        network_info: DockerNetworkInfo,
        certificate_volume_name: str,
        attempt: int,
    ):
        return self.create_child_task_with_common_params(
            SpawnTestContainer,
            test_container_name=self.test_container_name,
            network_info=network_info,
            ip_address_index_in_subnet=1,
            certificate_volume_name=certificate_volume_name,
            attempt=attempt,
            test_container_content=self.test_container_content,
        )

    def _wait_for_database(
        self, database_info: DatabaseInfo, attempt: int
    ) -> Generator[
        Union[WaitForTestDockerDatabase, WaitForTestExternalDatabase], None, bool
    ]:
        database_ready_target_future = yield from self.run_dependencies(
            self.create_wait_for_database_task(attempt, database_info)
        )
        is_database_ready: bool = self.get_values_from_futures(
            database_ready_target_future
        )
        return is_database_ready

    def create_wait_for_database_task(
        self, attempt: int, database_info: DatabaseInfo
    ) -> Union[WaitForTestDockerDatabase, WaitForTestExternalDatabase]:
        raise AbstractMethodException()



import logging
import os
from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    Callable,
    Optional,
)

import luigi
import networkx
from networkx.classes import DiGraph

from exasol_integration_test_docker_environment.lib.base.dependency_logger_base_task import (
    DependencyLoggerBaseTask,
)
from exasol_integration_test_docker_environment.lib.base.task_dependency import (
    DependencyState,
    TaskDependency,
)
from exasol_integration_test_docker_environment.lib.logging.configure_logging import (
    configure_logging,
)
from exasol_integration_test_docker_environment.lib.logging.luigi_log_config import (
    get_log_path,
)
from exasol_integration_test_docker_environment.lib.models.api_errors import (
    TaskFailures,
    TaskRuntimeError,
)
from exasol_integration_test_docker_environment.lib.utils.job_counter_singleton import (
    JobCounterSingleton,
)


def generate_root_task(task_class, *args, **kwargs) -> DependencyLoggerBaseTask:
    job_counter = JobCounterSingleton().get_next_value()
    strftime = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
    params = {"job_id": f"{strftime}_{job_counter}_{task_class.__name__}"}
    params.update(kwargs)
    return task_class(**params)


def run_task(
    task_creator: Callable[[], DependencyLoggerBaseTask],
    workers: int = 2,
    task_dependencies_dot_file: Optional[str] = None,
    log_level: Optional[str] = None,
    use_job_specific_log_file: bool = False,
) -> Any:
    setup_worker()
    task = task_creator()
    success = False
    log_file_path = get_log_path(task.job_id)
    try:
        no_scheduling_errors = _run_task_with_logging_config(
            task, log_file_path, log_level, use_job_specific_log_file, workers
        )
        success = not task.failed_target.exists() and no_scheduling_errors
        return _handle_task_result(
            no_scheduling_errors, success, task, task_dependencies_dot_file
        )
    except BaseException as e:
        logging.error("Going to abort the task %s" % task)
        raise e
    finally:
        if use_job_specific_log_file:
            logging.info(
                f"The detailed log of the integration-test-docker-environment can be found at: {log_file_path}"
            )
        task.cleanup(success)


def _run_task_with_logging_config(
    task: DependencyLoggerBaseTask,
    log_file_path: Path,
    log_level: Optional[str],
    use_job_specific_log_file: bool,
    workers: int,
) -> bool:
    with configure_logging(
        log_file_path, log_level, use_job_specific_log_file
    ) as run_kwargs:
        no_scheduling_errors = luigi.build(
            [task], workers=workers, local_scheduler=True, **run_kwargs
        )
        return no_scheduling_errors


def _handle_task_result(
    no_scheduling_errors: bool,
    success: bool,
    task: DependencyLoggerBaseTask,
    task_dependencies_dot_file: Optional[str],
) -> Any:
    generate_graph_from_task_dependencies(task, task_dependencies_dot_file)
    if success:
        return task.get_result()
    elif task.failed_target.exists():
        logging.error(f"Task {task} failed. failed target exists.")
        task_failures = list(task.collect_failures().keys())
        raise TaskRuntimeError(
            msg=f"Task {task} (or any of it's subtasks) failed.", inner=task_failures
        ) from TaskFailures(inner=task_failures)
    elif not no_scheduling_errors:
        logging.error(f"Task {task} failed. : luigi reported a scheduling error.")
        raise TaskRuntimeError(
            msg=f"Task {task} failed. reason: luigi reported a scheduling error."
        )


def generate_graph_from_task_dependencies(
    task: DependencyLoggerBaseTask, task_dependencies_dot_file: Optional[str]
):
    if task_dependencies_dot_file is not None:
        print(f"Generate Task Dependency Graph to {task_dependencies_dot_file}")
        print()
        dependencies = collect_dependencies(task)
        g = DiGraph()
        for dependency in dependencies:
            g.add_node(
                dependency.source.formatted_id,
                label=dependency.source.formatted_representation,
            )
            g.add_node(
                dependency.target.formatted_id,
                label=dependency.target.formatted_representation,
            )
            g.add_edge(
                dependency.source.formatted_id,
                dependency.target.formatted_id,
                dependency=dependency.formatted,
                label=f'"type={dependency.type}, index={dependency.index}"',
            )
        networkx.nx_pydot.write_dot(g, task_dependencies_dot_file)


def collect_dependencies(task: DependencyLoggerBaseTask) -> set[TaskDependency]:
    dependencies: set[TaskDependency] = set()
    for root, directories, files in os.walk(task._get_dependencies_path_for_job()):
        for file in files:
            file_path = Path(root).joinpath(file)
            with open(file_path) as f:
                for line in f.readlines():
                    task_dependency: TaskDependency = TaskDependency.from_json(line)  # type: ignore
                    if task_dependency.state == DependencyState.requested.name:
                        dependencies.add(task_dependency)
    return dependencies


def setup_worker():
    luigi.configuration.get_config().set("worker", "wait_interval", str(0.1))
    luigi.configuration.get_config().set("worker", "wait_jitter", str(0.5))



def cli_function(func):
    """Decorator: Register a function as having a cli equivalent"""
    func.__cli_function__ = True
    return func


def no_cli_function(func):
    """Decorator: Register a function as not having a cli equivalent"""
    func.__cli_function__ = False
    return func



class JobCounterSingleton:
    """
    We use here a Singleton to avoid an unprotected global variable.
    However, this counter needs to be global counter to guarantee unique job ids.
    This is needed in case of task that finish in less than a second, to avoid duplicated job ids
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._counter = 0
        return cls._instance

    def get_next_value(self) -> int:
        # self._counter is a class variable and because of this we need to suppress type checks
        self._counter += 1  # type: ignore
        return self._counter  # type: ignore



import click

DEFAULT_DOCKER_REPOSITORY_NAME = "exasol/script-language-container"

docker_repository_options = [
    click.option(
        "--source-docker-repository-name",
        type=str,
        default=DEFAULT_DOCKER_REPOSITORY_NAME,
        show_default=True,
        help="Name of the docker repository for pulling cached stages. "
        "The repository name may contain the URL of the docker registry, "
        "the username and the actual repository name. "
        "A common structure is <docker-registry-url>/<username>/<repository-name>",
    ),
    click.option(
        "--source-docker-tag-prefix",
        type=str,
        default="",
        show_default=True,
        help="Prefix for the tags which are used for pulling of cached stages",
    ),
    click.option(
        "--source-docker-username",
        type=str,
        default=None,
        help="Username for the docker registry from where the system pulls cached stages.",
        required=False,
    ),
    click.option(
        "--source-docker-password",
        type=str,
        default=None,
        help="Password for the docker registry from where the system pulls cached stages. "
        "Without password option the system prompts for the password.",
        required=False,
    ),
    click.option(
        "--target-docker-repository-name",
        type=str,
        default=DEFAULT_DOCKER_REPOSITORY_NAME,
        show_default=True,
        help="Name of the docker repository for naming and pushing images of stages. "
        "The repository name may contain the URL of the docker registry, "
        "the username and the actual repository name. "
        "A common structure is <docker-registry-url>/<username>/<repository-name>",
    ),
    click.option(
        "--target-docker-tag-prefix",
        type=str,
        default="",
        show_default=True,
        help="Prefix for the tags which are used for naming and pushing of stages",
    ),
    click.option(
        "--target-docker-username",
        type=str,
        default=None,
        help="Username for the docker registry where the system pushes images of stages.",
        required=False,
    ),
    click.option(
        "--target-docker-password",
        type=str,
        default=None,
        help="Password for the docker registry where the system pushes images of stages. "
        "Without password option the system prompts for the password.",
        required=False,
    ),
]

simple_docker_repository_options = [
    click.option(
        "--docker-repository-name",
        type=str,
        default=DEFAULT_DOCKER_REPOSITORY_NAME,
        show_default=True,
        help="Name of the docker repository for naming images. "
        "The repository name may contain the URL of the docker registry, "
        "the username and the actual repository name. "
        "A common structure is <docker-registry-url>/<username>/<repository-name>",
    ),
    click.option(
        "--docker-tag-prefix",
        type=str,
        default="",
        show_default=True,
        help="Prefix for the tags of the images",
    ),
]



from pathlib import Path
from typing import (
    Optional,
)

from exasol_integration_test_docker_environment.lib.base.info import Info


class TestContainerBuildMapping(Info):
    """
    Represents a mapping of a build artifact for the test-container.
    The artifact will be copied to location "target", parallel to the Dockerfile and is hence accessible
    from within the Dockerfile during the build time of the test-container.
    """

    def __init__(self, source: Path, target: str) -> None:
        self.source = source
        self.target = target


class TestContainerRuntimeMapping(Info):
    """
    Represents a mapping of a runtime artifact for the test-container.
    This artifact will be bind-mounted to the test-container during startup.
    Optionally, the content will be copied within the test-container to the location indicated by parameter
    "deployement_target": This is useful if the source path must not be polluted with runtime artifacts (logs, etc.).
    """

    def __init__(
        self, source: Path, target: str, deployment_target: Optional[str] = None
    ) -> None:
        self.source = source
        self.target = target
        self.deployment_target = deployment_target


class TestContainerContentDescription(Info):
    """
    This class contains all information necessary to build, start and set up the test-container.
    Its purpose is to give the client control about the build- and runtime-artifacts.
    """

    def __init__(
        self,
        docker_file: Optional[str],
        build_files_and_directories: list[TestContainerBuildMapping],
        runtime_mappings: list[TestContainerRuntimeMapping],
    ) -> None:
        self.docker_file = docker_file
        self.build_files_and_directories = build_files_and_directories
        self.runtime_mappings = runtime_mappings



import luigi
from luigi import Config
from luigi.parameter import ParameterVisibility


class DockerParameter(Config):
    """
    Docker parameters used for Tasks accessing Docker client.
    """

    timeout: int = luigi.IntParameter(
        default=100000, significant=False, visibility=ParameterVisibility.PRIVATE
    )
    no_cache: bool = luigi.BoolParameter(default=False)



import functools
import warnings
from typing import (
    Any,
    Callable,
    Optional,
)

import humanfriendly

from exasol_integration_test_docker_environment.cli.options.docker_repository_options import (
    DEFAULT_DOCKER_REPOSITORY_NAME,
)
from exasol_integration_test_docker_environment.cli.options.system_options import (
    DEFAULT_OUTPUT_DIRECTORY,
)
from exasol_integration_test_docker_environment.cli.options.test_environment_options import (
    LATEST_DB_VERSION,
)
from exasol_integration_test_docker_environment.lib.base.run_task import (
    generate_root_task,
    run_task,
)
from exasol_integration_test_docker_environment.lib.docker.container.utils import (
    remove_docker_container,
)
from exasol_integration_test_docker_environment.lib.docker.networks.utils import (
    remove_docker_networks,
)
from exasol_integration_test_docker_environment.lib.docker.volumes.utils import (
    remove_docker_volumes,
)
from exasol_integration_test_docker_environment.lib.models.api_errors import (
    ArgumentConstraintError,
)
from exasol_integration_test_docker_environment.lib.models.config.build_config import (
    set_build_config,
)
from exasol_integration_test_docker_environment.lib.models.config.docker_config import (
    set_docker_repository_config,
)
from exasol_integration_test_docker_environment.lib.models.data.environment_info import (
    EnvironmentInfo,
)
from exasol_integration_test_docker_environment.lib.test_environment.parameter.docker_db_test_environment_parameter import (
    DbOsAccess,
)
from exasol_integration_test_docker_environment.lib.test_environment.spawn_test_environment_with_docker_db import (
    SpawnTestEnvironmentWithDockerDB,
)
from exasol_integration_test_docker_environment.lib.utils.api_function_decorators import (
    cli_function,
)


def _cleanup(environment_info: EnvironmentInfo) -> None:
    if environment_info.database_info.container_info is not None:
        remove_docker_container(
            [environment_info.database_info.container_info.container_name]
        )
        if environment_info.database_info.container_info.volume_name is not None:
            remove_docker_volumes(
                [environment_info.database_info.container_info.volume_name]
            )
    remove_docker_networks([environment_info.network_info.network_name])


@cli_function
def spawn_test_environment(
    environment_name: str,
    database_port_forward: Optional[int] = None,
    bucketfs_port_forward: Optional[int] = None,
    ssh_port_forward: Optional[int] = None,
    db_mem_size: str = "2 GiB",
    db_disk_size: str = "2 GiB",
    nameserver: tuple[str, ...] = (),
    docker_runtime: Optional[str] = None,
    docker_db_image_version: str = LATEST_DB_VERSION,
    docker_db_image_name: str = "exasol/docker-db",
    db_os_access: Optional[str] = "DOCKER_EXEC",
    create_certificates: bool = False,
    additional_db_parameter: tuple[str, ...] = (),
    docker_environment_variable: tuple[str, ...] = (),
    accelerator: tuple[str, ...] = (),
    source_docker_repository_name: str = DEFAULT_DOCKER_REPOSITORY_NAME,
    source_docker_tag_prefix: str = "",
    source_docker_username: Optional[str] = None,
    source_docker_password: Optional[str] = None,
    target_docker_repository_name: str = DEFAULT_DOCKER_REPOSITORY_NAME,
    target_docker_tag_prefix: str = "",
    target_docker_username: Optional[str] = None,
    target_docker_password: Optional[str] = None,
    output_directory: str = DEFAULT_OUTPUT_DIRECTORY,
    temporary_base_directory: str = "/tmp",
    workers: int = 5,
    task_dependencies_dot_file: Optional[str] = None,
    log_level: Optional[str] = None,
    use_job_specific_log_file: bool = False,
    bucketfs_http_port_forward: Optional[int] = None,
    bucketfs_https_port_forward: Optional[int] = None,
) -> tuple[EnvironmentInfo, Callable[[], None]]:
    """
    This function spawns a test environment with a docker-db container and a connected test-container.
    The test-container is reachable by the database for output redirects of UDFs.
    The function returns an environment_info object, describing the environment, and a cleanup-method, which
    can be used to stop the environment.
    raises: TaskRuntimeError if spawning the test environment fails

    """

    def str_or_none(x: Any) -> Optional[str]:
        return str(x) if x is not None else None

    parsed_db_mem_size = humanfriendly.parse_size(db_mem_size)
    if parsed_db_mem_size < humanfriendly.parse_size("1 GiB"):
        raise ArgumentConstraintError("db_mem_size", "needs to be at least 1 GiB")
    parsed_db_disk_size = humanfriendly.parse_size(db_disk_size)
    if parsed_db_disk_size < humanfriendly.parse_size("100 MiB"):
        raise ArgumentConstraintError("db_disk_size", "needs to be at least 100 MiB")
    db_os_access_value = (
        DbOsAccess[db_os_access] if db_os_access else DbOsAccess.DOCKER_EXEC
    )
    if len(accelerator) > 0 and accelerator != ("nvidia=all",):
        raise ArgumentConstraintError(
            "accelerator", "Only value 'nvidia=all' is supported"
        )
    if bucketfs_port_forward is not None:
        warnings.warn(
            "Parameter 'bucketfs_port_forward' is deprecated. Use 'bucketfs_http_port_forward' instead.",
            DeprecationWarning,
        )

    set_build_config(
        False,
        (),
        False,
        False,
        output_directory,
        temporary_base_directory,
        None,
        None,
    )
    set_docker_repository_config(
        source_docker_password,
        source_docker_repository_name,
        source_docker_username,
        source_docker_tag_prefix,
        "source",
    )
    set_docker_repository_config(
        target_docker_password,
        target_docker_repository_name,
        target_docker_username,
        target_docker_tag_prefix,
        "target",
    )
    task_creator = lambda: generate_root_task(
        task_class=SpawnTestEnvironmentWithDockerDB,
        environment_name=environment_name,
        database_port_forward=str_or_none(database_port_forward),
        bucketfs_port_forward=str_or_none(bucketfs_port_forward),
        bucketfs_http_port_forward=str_or_none(bucketfs_port_forward),
        bucketfs_https_port_forward=str_or_none(bucketfs_https_port_forward),
        ssh_port_forward=str_or_none(ssh_port_forward),
        mem_size=db_mem_size,
        disk_size=db_disk_size,
        nameservers=nameserver,
        docker_runtime=docker_runtime,
        docker_db_image_version=docker_db_image_version,
        docker_db_image_name=docker_db_image_name,
        db_os_access=db_os_access_value,
        db_user="sys",
        db_password="exasol",
        bucketfs_write_password="write",
        no_test_container_cleanup_after_success=True,
        no_test_container_cleanup_after_failure=False,
        no_database_cleanup_after_success=True,
        no_database_cleanup_after_failure=False,
        create_certificates=create_certificates,
        test_container_content=None,
        additional_db_parameter=additional_db_parameter,
        docker_environment_variables=docker_environment_variable,
        accelerator=accelerator,
    )
    environment_info = run_task(
        task_creator,
        workers,
        task_dependencies_dot_file,
        log_level=log_level,
        use_job_specific_log_file=use_job_specific_log_file,
    )
    return environment_info, functools.partial(_cleanup, environment_info)



import hashlib
import os
import stat
from dataclasses import dataclass
from multiprocessing import Pool
from pathlib import (
    Path,
    PurePath,
)
from typing import (
    Callable,
)

import humanfriendly

from exasol_integration_test_docker_environment.lib.docker.images.create.utils.character_length_checker import (
    CharacterLengthChecker,
)
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.symlink_loop_checker import (
    SymlinkLoopChecker,
)

HASH_FUNCTIONS = {
    "md5": hashlib.md5,
    "sha1": hashlib.sha1,
    "sha256": hashlib.sha256,
    "sha512": hashlib.sha512,
}


@dataclass(frozen=True)
class PathMapping:
    """
    Describes a mapping of a file or directory from source to destination.
    """

    destination: PurePath
    source: Path


@dataclass(frozen=True)
class DestinationMapping:
    """
    Represents one file/directory found by traversing the source directory tree.
    """

    destination_path: PurePath
    """The path of the file/directory in the destination (including destination_root_path)"""
    source_path: Path
    """The path of the file/directory in the source"""

    def use_for_hashing(self, hash_files: bool, hash_directories: bool) -> bool:
        """
        Returns if this file/directory entry should be used for file hashing, considering parameters
        "hash_files" and "hash_directories".
        """
        if self.source_path.is_file():
            return hash_files
        else:
            return hash_directories


@dataclass(frozen=True)
class DirectoryMappingResult:
    """
    Contains all entries found by traversing all mapping directory paths.
    """

    sources: list[Path]
    paths_for_hashing: list[DestinationMapping]


class FileDirectoryListHasher:

    def __init__(
        self,
        hashfunc: str = "md5",
        followlinks: bool = False,
        hash_file_names: bool = False,
        hash_permissions: bool = False,
        hash_directory_names: bool = False,
        excluded_directories=None,
        excluded_files=None,
        excluded_extensions=None,
        blocksize: str = "64kb",
        workers: int = 4,
        max_characters_paths: int = 500000000,
    ) -> None:
        self.MAX_CHARACTERS_PATHS = max_characters_paths
        self.workers = workers
        self.excluded_files = excluded_files
        self.excluded_extensions = excluded_extensions
        self.excluded_directories = excluded_directories
        self.hash_directory_names = hash_directory_names
        self.hash_permissions = hash_permissions
        self.hash_file_names = hash_file_names
        self.followlinks = followlinks

        self.hash_func = HASH_FUNCTIONS.get(hashfunc)
        if not self.hash_func:
            raise NotImplementedError(f"{hashfunc} not implemented.")

        if not self.excluded_files:
            self.excluded_files = []

        if not self.excluded_directories:
            self.excluded_directories = []

        if not self.excluded_extensions:
            self.excluded_extensions = []

        self.path_hasher = PathHasher(hashfunc, hash_permissions=hash_permissions)
        self.file_content_hasher = FileContentHasher(hashfunc, blocksize)

    def hash(self, files_and_directories: list[PathMapping]) -> bytes:
        if not isinstance(files_and_directories, list):
            raise Exception(
                "List with paths expected and not '%s' with type %s"
                % (files_and_directories, type(files_and_directories))
            )

        directory_mapping_result = self.collect_dest_path_and_src_files(
            files_and_directories
        )
        hashes = self.compute_hashes(directory_mapping_result)
        return self._reduce_hash(hashes)

    @staticmethod
    def check_no_duplicate_destinations(mappings: list[DestinationMapping]) -> None:
        """
        Verify that there are no duplicate mappings to same destination.
        This can happen, if the destination of two mappings are equal and the two sources contains the same sub-path
        structure; or if subpaths of destinations and sources somehow match;
        or if the destination of two mappings is the same file.
        """
        destination_paths_set = {p.destination_path for p in mappings}
        if len(destination_paths_set) != len(mappings):
            raise AssertionError(
                f"Directory content for hashing contains duplicates: {mappings}"
            )

    def collect_dest_path_and_src_files(
        self, files_and_directories: list[PathMapping]
    ) -> DirectoryMappingResult:
        """
        Traverse the source paths of all mappings and assemble two lists:
        1) Destination paths-names (for Hashing):
            Mapping of source file/directory -> destination of file/directory,
            ordered by destination path. List[DestinationMapping].
        2) Source files, ordered by destination path
        Collection is done considering the current configuration
         - for both: excluded_directory/excluded_extensions/excluded_file
         - for 1): hash_file_names/hash_directory_names
        """
        collected_dest_paths: list[DestinationMapping] = []

        def replace_src_by_dest_path(
            src: PurePath, dest: PurePath, target: str
        ) -> PurePath:
            if not target.startswith(str(src)):
                raise RuntimeError(
                    f"path target {target} does not start with source: {src}"
                )
            p = Path(target).relative_to(src)
            return Path(dest) / p

        for file_or_directory in files_and_directories:
            source = file_or_directory.source
            destination = file_or_directory.destination

            def handle_directory(directories: list[str]) -> None:
                new_dest_paths_mappings = [
                    DestinationMapping(
                        destination_path=replace_src_by_dest_path(
                            source, destination, p
                        ),
                        source_path=Path(p),
                    )
                    for p in directories
                ]
                collected_dest_paths.extend(new_dest_paths_mappings)

            def handle_files(files: list[str]) -> None:
                collected_dest_paths.extend(
                    [
                        DestinationMapping(
                            destination_path=replace_src_by_dest_path(
                                source, destination, f
                            ),
                            source_path=Path(f),
                        )
                        for f in files
                    ]
                )

            if os.path.isdir(source):
                self.traverse_directory(source, handle_directory, handle_files)
            elif os.path.isfile(source):
                collected_dest_paths.append(
                    DestinationMapping(
                        destination_path=Path(destination), source_path=Path(source)
                    )
                )
            else:
                raise FileNotFoundError("Could not find file or directory %s" % source)
        collected_dest_paths.sort(key=lambda x: x.destination_path)
        self.check_no_duplicate_destinations(collected_dest_paths)

        # Now, after we sorted the collected paths, filter out paths which are not needed for current configuration
        filtered_dest_paths = [
            d
            for d in collected_dest_paths
            if d.use_for_hashing(self.hash_file_names, self.hash_directory_names)
        ]

        collected_src_files = [
            p.source_path for p in collected_dest_paths if p.source_path.is_file()
        ]
        return DirectoryMappingResult(
            sources=collected_src_files, paths_for_hashing=filtered_dest_paths
        )

    def compute_hashes(
        self, directory_mapping_result: DirectoryMappingResult
    ) -> list[bytes]:
        paths_for_hashing = directory_mapping_result.paths_for_hashing
        collected_src_files = directory_mapping_result.sources
        pool = Pool(processes=self.workers)
        dest_path_hashes_future = pool.map_async(
            self.path_hasher.hash, paths_for_hashing, chunksize=2
        )
        file_content_hashes_future = pool.map_async(
            self.file_content_hasher.hash, collected_src_files, chunksize=2
        )
        hashes = []
        hashes.extend(file_content_hashes_future.get())
        file_path_hashes_of_directories = dest_path_hashes_future.get()
        hashes.extend(file_path_hashes_of_directories)
        return hashes

    def has_excluded_extension(self, f: str):
        return f.split(".")[-1:][0] in self.excluded_extensions

    def is_excluded_file(self, f: str):
        return f in self.excluded_files

    def is_excluded_directory(self, f: str):
        return f in self.excluded_directories

    def traverse_directory(
        self,
        directory: PurePath,
        directory_handler: Callable[[list[str]], None],
        file_handler: Callable[[list[str]], None],
    ) -> None:

        symlink_loop_checker = SymlinkLoopChecker()
        character_length_checker = CharacterLengthChecker(
            directory,
            self.MAX_CHARACTERS_PATHS,
            self.hash_directory_names,
            self.hash_file_names,
        )

        for root, dirs, files in os.walk(
            directory, topdown=True, followlinks=self.followlinks
        ):
            symlink_loop_checker.check_and_add(root)

            new_directories = [
                os.path.join(root, d) for d in dirs if not self.is_excluded_directory(d)
            ]
            directory_handler(new_directories)
            new_files = [
                os.path.join(root, f)
                for f in files
                if not self.is_excluded_file(f) and not self.has_excluded_extension(f)
            ]
            character_length_checker.add_and_check(new_directories, new_files)
            file_handler(new_files)

    def _reduce_hash(self, hashes):
        hasher = self.hash_func()
        for hashvalue in hashes:
            hasher.update(hashvalue)
        return hasher.digest()


class PathHasher:
    def __init__(self, hashfunc: str = "md5", hash_permissions: bool = False) -> None:
        self.hash_permissions = hash_permissions
        self.hash_func = HASH_FUNCTIONS.get(hashfunc)
        if not self.hash_func:
            raise NotImplementedError(f"{hashfunc} not implemented.")

    def hash(self, path_mapping: DestinationMapping) -> bytes:
        src_path = path_mapping.source_path
        dest_path = path_mapping.destination_path
        assert self.hash_func
        hasher = self.hash_func()
        hasher.update(str(dest_path).encode("utf-8"))
        if self.hash_permissions:
            stat_result = os.stat(src_path)
            # we only check the executable right of the user, because git only remembers this
            user_has_executable_rights = stat.S_IXUSR & stat_result[stat.ST_MODE]
            hasher.update(str(user_has_executable_rights).encode("utf-8"))
        return hasher.digest()


class FileContentHasher:

    def __init__(self, hashfunc: str = "md5", blocksize: str = "64kb") -> None:
        self.blocksize = humanfriendly.parse_size(blocksize)
        self.hash_func = HASH_FUNCTIONS.get(hashfunc)
        if not self.hash_func:
            raise NotImplementedError(f"{hashfunc} not implemented.")

    def hash(self, filepath: Path) -> bytes:
        assert self.hash_func
        hasher = self.hash_func()
        with open(filepath, "rb") as fp:
            while True:
                data = fp.read(self.blocksize)
                if not data:
                    break
                hasher.update(data)
        return hasher.digest()



import pickle
from pathlib import Path

from luigi import LocalTarget
from luigi.format import NopFormat

FORMAT = NopFormat()


class PickleTarget(LocalTarget):

    def __init__(self, path: Path, is_tmp: bool = False) -> None:
        super().__init__(path=str(path), is_tmp=is_tmp, format=FORMAT)

    def write(self, obj) -> None:
        with self.open("w") as f:
            pickle.dump(obj, f)

    def read(self):
        with self.open("r") as f:
            obj = pickle.load(f)
            return obj
